from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict

from agno.infra.settings import InfraSettings


class InfraBase(BaseModel):
    """Base class for all InfraResource, InfraApp and InfraResources objects."""

    # Name of the infrastructure resource
    name: Optional[str] = None
    # Group for the infrastructure resource
    # Used for filtering infrastructure resources by group
    group: Optional[str] = None
    # Environment filter for this resource
    env: Optional[str] = None
    # Infrastructure filter for this resource
    infra: Optional[str] = None
    # Whether this resource is enabled
    enabled: bool = True

    # Resource Control
    skip_create: bool = False
    skip_read: bool = False
    skip_update: bool = False
    skip_delete: bool = False
    recreate_on_update: bool = False

    # Skip create if resource with the same name is active
    use_cache: bool = True

    # Force create/update/delete even if a resource with the same name is active
    force: Optional[bool] = None

    # Wait for resource to be created, updated or deleted
    wait_for_create: bool = True
    wait_for_update: bool = True
    wait_for_delete: bool = True
    waiter_delay: int = 30
    waiter_max_attempts: int = 50

    # Environment Variables for the resource (if applicable)
    # Add env variables to resource where applicable
    env_vars: Optional[Dict[str, Any]] = None
    # Read env from a file in yaml format
    env_file: Optional[Path] = None

    # Add secret variables to resource where applicable
    # secrets_dict: Optional[Dict[str, Any]] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Debug Mode
    debug_mode: bool = False

    # Store resource to output directory
    # If True, save resource output to json files
    save_output: bool = False

    # The directory for the input files in the infra directory
    input_dir: Optional[str] = None
    # The directory for the output files in the infra directory
    output_dir: Optional[str] = None

    # Dependencies for the resource
    depends_on: Optional[List[Any]] = None

    # Infra Settings
    infra_settings: Optional[InfraSettings] = None

    # Cached Data
    cached_infra_dir: Optional[Path] = None
    cached_env_file_data: Optional[Dict[str, Any]] = None
    cached_secret_file_data: Optional[Dict[str, Any]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def get_group_name(self) -> Optional[str]:
        return self.group or self.name

    @property
    def infra_root(self) -> Optional[Path]:
        return self.infra_settings.infra_root if self.infra_settings is not None else None

    @property
    def infra_name(self) -> Optional[str]:
        return self.infra_settings.infra_name if self.infra_settings is not None else None

    @property
    def infra_dir(self) -> Optional[Path]:
        if self.cached_infra_dir is not None:
            return self.cached_infra_dir

        if self.infra_root is not None:
            from agno.infra.helpers import get_infra_dir_path

            infra_dir = get_infra_dir_path(self.infra_root)
            if infra_dir is not None:
                self.cached_infra_dir = infra_dir
                return infra_dir
        return None

    def set_infra_settings(self, infra_settings: Optional[InfraSettings] = None) -> None:
        if infra_settings is not None:
            self.infra_settings = infra_settings

    def get_env_file_data(self) -> Optional[Dict[str, Any]]:
        if self.cached_env_file_data is None:
            from agno.utilities.yaml_io import read_yaml_file

            self.cached_env_file_data = read_yaml_file(file_path=self.env_file)
        return self.cached_env_file_data

    def get_secret_file_data(self) -> Optional[Dict[str, Any]]:
        if self.cached_secret_file_data is None:
            from agno.utilities.yaml_io import read_yaml_file

            self.cached_secret_file_data = read_yaml_file(file_path=self.secrets_file)
        return self.cached_secret_file_data

    def get_secret_from_file(self, secret_name: str) -> Optional[str]:
        secret_file_data = self.get_secret_file_data()
        if secret_file_data is not None:
            return secret_file_data.get(secret_name)
        return None

    def get_infra_resources(self) -> Optional[Any]:
        """This method returns an InfraResources object for this resource"""
        raise NotImplementedError("get_infra_resources method not implemented")



from typing import Optional

from agno.aws.api_client import AwsApiClient


class AwsReference:
    def __init__(self, reference):
        self.reference = reference

    def get_reference(self, aws_client: Optional[AwsApiClient] = None):
        return self.reference(aws_client=aws_client)



from typing import Any, List, Optional

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.cli.console import print_info
from agno.utilities.logging import logger


class CloudFormationStack(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#service-resource
    """

    resource_type: Optional[str] = "CloudFormationStack"
    service_name: str = "cloudformation"

    # StackName: The name must be unique in the Region in which you are creating the stack.
    name: str
    # Location of file containing the template body.
    # The URL must point to a template (max size: 460,800 bytes) that's located in an
    # Amazon S3 bucket or a Systems Manager document.
    template_url: str
    # parameters: Optional[List[Dict[str, Union[str, bool]]]] = None
    # disable_rollback: Optional[bool] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Create CloudFormationStack
        service_resource = self.get_service_resource(aws_client)
        try:
            stack = service_resource.create_stack(
                StackName=self.name,
                TemplateURL=self.template_url,
            )
            logger.debug(f"Stack: {stack}")

            # Validate Stack creation
            stack.load()
            creation_time = stack.creation_time
            logger.debug(f"creation_time: {creation_time}")
            if creation_time is not None:
                self.active_resource = stack
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Stack to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                waiter = self.get_service_client(aws_client).get_waiter("stack_create_complete")
                waiter.wait(
                    StackName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
                return False
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_resource = self.get_service_resource(aws_client)
        try:
            stack = service_resource.Stack(name=self.name)

            stack.load()
            creation_time = stack.creation_time
            logger.debug(f"creation_time: {creation_time}")
            if creation_time is not None:
                logger.debug(f"Stack found: {stack.stack_name}")
                self.active_resource = stack
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        self.active_resource = None
        try:
            stack = self._read(aws_client)
            logger.debug(f"Stack: {stack}")
            if stack is None:
                logger.warning(f"No {self.get_resource_type()} to delete")
                return True

            stack.delete()
            # print_info("Stack deleted")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for Stack to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("stack_delete_complete")
                waiter.wait(
                    StackName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
                return True
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def get_stack_resource(self, aws_client: AwsApiClient, logical_id: str) -> Optional[Any]:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#CloudFormation.StackResource
        # logger.debug(f"Getting StackResource {logical_id} for {self.name}")
        try:
            service_resource = self.get_service_resource(aws_client)
            stack_resource = service_resource.StackResource(self.name, logical_id)
            return stack_resource
        except Exception as e:
            logger.error(e)
        return None

    def get_stack_resource_physical_id(self, stack_resource: Any) -> Optional[str]:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#CloudFormation.StackResource
        try:
            physical_resource_id = stack_resource.physical_resource_id if stack_resource is not None else None
            logger.debug(f"{stack_resource.logical_id}: {physical_resource_id}")
            return physical_resource_id
        except Exception:
            return None

    def get_private_subnets(self, aws_client: Optional[AwsApiClient] = None) -> Optional[List[str]]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile())
            )

            private_subnets = []

            private_subnet_1_stack_resource = self.get_stack_resource(client, "PrivateSubnet01")
            private_subnet_1_physical_resource_id = self.get_stack_resource_physical_id(private_subnet_1_stack_resource)
            if private_subnet_1_physical_resource_id is not None:
                private_subnets.append(private_subnet_1_physical_resource_id)

            private_subnet_2_stack_resource = self.get_stack_resource(client, "PrivateSubnet02")
            private_subnet_2_physical_resource_id = self.get_stack_resource_physical_id(private_subnet_2_stack_resource)
            if private_subnet_2_physical_resource_id is not None:
                private_subnets.append(private_subnet_2_physical_resource_id)

            private_subnet_3_stack_resource = self.get_stack_resource(client, "PrivateSubnet03")
            private_subnet_3_physical_resource_id = self.get_stack_resource_physical_id(private_subnet_3_stack_resource)
            if private_subnet_3_physical_resource_id is not None:
                private_subnets.append(private_subnet_3_physical_resource_id)

            return private_subnets if (len(private_subnets) > 0) else None
        except Exception as e:
            logger.error(e)
        return None

    def get_public_subnets(self, aws_client: Optional[AwsApiClient] = None) -> Optional[List[str]]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile())
            )

            public_subnets = []

            public_subnet_1_stack_resource = self.get_stack_resource(client, "PublicSubnet01")
            public_subnet_1_physical_resource_id = self.get_stack_resource_physical_id(public_subnet_1_stack_resource)
            if public_subnet_1_physical_resource_id is not None:
                public_subnets.append(public_subnet_1_physical_resource_id)

            public_subnet_2_stack_resource = self.get_stack_resource(client, "PublicSubnet02")
            public_subnet_2_physical_resource_id = self.get_stack_resource_physical_id(public_subnet_2_stack_resource)
            if public_subnet_2_physical_resource_id is not None:
                public_subnets.append(public_subnet_2_physical_resource_id)

            return public_subnets if (len(public_subnets) > 0) else None
        except Exception as e:
            logger.error(e)
        return None

    def get_security_group(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile())
            )

            security_group_stack_resource = self.get_stack_resource(client, "ControlPlaneSecurityGroup")
            security_group_physical_resource_id = (
                security_group_stack_resource.physical_resource_id
                if security_group_stack_resource is not None
                else None
            )
            logger.debug(f"security_group: {security_group_physical_resource_id}")

            return security_group_physical_resource_id
        except Exception as e:
            logger.error(e)
        return None



from typing import Any, Dict, List, Optional, Union

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.aws.resource.secret.manager import SecretsManager
from agno.aws.resource.secret.reader import read_secrets
from agno.utilities.logging import logger


class EcsContainer(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html
    """

    resource_type: Optional[str] = "EcsContainer"
    service_name: str = "ecs"

    # The name of a container.
    # If you're linking multiple containers together in a task definition, the name of one container can be entered in
    # the links of another container to connect the containers.
    name: str
    # The image used to start a container.
    image: str
    # The private repository authentication credentials to use.
    repository_credentials: Optional[Dict[str, Any]] = None
    # The number of cpu units reserved for the container.
    cpu: Optional[int] = None
    # The amount (in MiB) of memory to present to the container.
    memory: Optional[int] = None
    # The soft limit (in MiB) of memory to reserve for the container.
    memory_reservation: Optional[int] = None
    # The links parameter allows containers to communicate with each other without the need for port mappings.
    links: Optional[List[str]] = None
    # The list of port mappings for the container. Port mappings allow containers to access ports on the host container
    # instance to send or receive traffic.
    port_mappings: Optional[List[Dict[str, Any]]] = None
    # If the essential parameter of a container is marked as true , and that container fails or stops for any reason,
    # all other containers that are part of the task are stopped. If the essential parameter of a container is marked
    # as false , its failure doesn't affect the rest of the containers in a task. If this parameter is omitted,
    # a container is assumed to be essential.
    essential: Optional[bool] = None
    # The entry point that's passed to the container.
    entry_point: Optional[List[str]] = None
    # The command that's passed to the container.
    command: Optional[List[str]] = None
    # The environment variables to pass to a container.
    environment: Optional[List[Dict[str, Any]]] = None
    # A list of files containing the environment variables to pass to a container.
    environment_files: Optional[List[Dict[str, Any]]] = None
    # Read environment variables from AWS Secrets.
    env_from_secrets: Optional[Union[SecretsManager, List[SecretsManager]]] = None
    # The mount points for data volumes in your container.
    mount_points: Optional[List[Dict[str, Any]]] = None
    # Data volumes to mount from another container.
    volumes_from: Optional[List[Dict[str, Any]]] = None
    # Linux-specific modifications that are applied to the container, such as Linux kernel capabilities.
    linux_parameters: Optional[Dict[str, Any]] = None
    # The secrets to pass to the container.
    secrets: Optional[List[Dict[str, Any]]] = None
    # The dependencies defined for container startup and shutdown.
    depends_on: Optional[List[Dict[str, Any]]] = None
    # Time duration (in seconds) to wait before giving up on resolving dependencies for a container.
    start_timeout: Optional[int] = None
    # Time duration (in seconds) to wait before the container is forcefully killed if it doesn't exit normally.
    stop_timeout: Optional[int] = None
    # The hostname to use for your container.
    hostname: Optional[str] = None
    # The user to use inside the container.
    user: Optional[str] = None
    # The working directory to run commands inside the container in.
    working_directory: Optional[str] = None
    # When this parameter is true, networking is disabled within the container.
    disable_networking: Optional[bool] = None
    # When this parameter is true, the container is given elevated privileges
    # on the host container instance (similar to the root user).
    privileged: Optional[bool] = None
    readonly_root_filesystem: Optional[bool] = None
    dns_servers: Optional[List[str]] = None
    dns_search_domains: Optional[List[str]] = None
    extra_hosts: Optional[List[Dict[str, Any]]] = None
    docker_security_options: Optional[List[str]] = None
    interactive: Optional[bool] = None
    pseudo_terminal: Optional[bool] = None
    docker_labels: Optional[Dict[str, Any]] = None
    ulimits: Optional[List[Dict[str, Any]]] = None
    log_configuration: Optional[Dict[str, Any]] = None
    health_check: Optional[Dict[str, Any]] = None
    system_controls: Optional[List[Dict[str, Any]]] = None
    resource_requirements: Optional[List[Dict[str, Any]]] = None
    firelens_configuration: Optional[Dict[str, Any]] = None

    def get_container_definition(self, aws_client: Optional[AwsApiClient] = None) -> Dict[str, Any]:
        container_definition: Dict[str, Any] = {}

        # Build container environment
        container_environment: List[Dict[str, Any]] = self.build_container_environment(aws_client=aws_client)
        if container_environment is not None:
            container_definition["environment"] = container_environment

        if self.name is not None:
            container_definition["name"] = self.name
        if self.image is not None:
            container_definition["image"] = self.image
        if self.repository_credentials is not None:
            container_definition["repositoryCredentials"] = self.repository_credentials
        if self.cpu is not None:
            container_definition["cpu"] = self.cpu
        if self.memory is not None:
            container_definition["memory"] = self.memory
        if self.memory_reservation is not None:
            container_definition["memoryReservation"] = self.memory_reservation
        if self.links is not None:
            container_definition["links"] = self.links
        if self.port_mappings is not None:
            container_definition["portMappings"] = self.port_mappings
        if self.essential is not None:
            container_definition["essential"] = self.essential
        if self.entry_point is not None:
            container_definition["entryPoint"] = self.entry_point
        if self.command is not None:
            container_definition["command"] = self.command
        if self.environment_files is not None:
            container_definition["environmentFiles"] = self.environment_files
        if self.mount_points is not None:
            container_definition["mountPoints"] = self.mount_points
        if self.volumes_from is not None:
            container_definition["volumesFrom"] = self.volumes_from
        if self.linux_parameters is not None:
            container_definition["linuxParameters"] = self.linux_parameters
        if self.secrets is not None:
            container_definition["secrets"] = self.secrets
        if self.depends_on is not None:
            container_definition["dependsOn"] = self.depends_on
        if self.start_timeout is not None:
            container_definition["startTimeout"] = self.start_timeout
        if self.stop_timeout is not None:
            container_definition["stopTimeout"] = self.stop_timeout
        if self.hostname is not None:
            container_definition["hostname"] = self.hostname
        if self.user is not None:
            container_definition["user"] = self.user
        if self.working_directory is not None:
            container_definition["workingDirectory"] = self.working_directory
        if self.disable_networking is not None:
            container_definition["disableNetworking"] = self.disable_networking
        if self.privileged is not None:
            container_definition["privileged"] = self.privileged
        if self.readonly_root_filesystem is not None:
            container_definition["readonlyRootFilesystem"] = self.readonly_root_filesystem
        if self.dns_servers is not None:
            container_definition["dnsServers"] = self.dns_servers
        if self.dns_search_domains is not None:
            container_definition["dnsSearchDomains"] = self.dns_search_domains
        if self.extra_hosts is not None:
            container_definition["extraHosts"] = self.extra_hosts
        if self.docker_security_options is not None:
            container_definition["dockerSecurityOptions"] = self.docker_security_options
        if self.interactive is not None:
            container_definition["interactive"] = self.interactive
        if self.pseudo_terminal is not None:
            container_definition["pseudoTerminal"] = self.pseudo_terminal
        if self.docker_labels is not None:
            container_definition["dockerLabels"] = self.docker_labels
        if self.ulimits is not None:
            container_definition["ulimits"] = self.ulimits
        if self.log_configuration is not None:
            container_definition["logConfiguration"] = self.log_configuration
        if self.health_check is not None:
            container_definition["healthCheck"] = self.health_check
        if self.system_controls is not None:
            container_definition["systemControls"] = self.system_controls
        if self.resource_requirements is not None:
            container_definition["resourceRequirements"] = self.resource_requirements
        if self.firelens_configuration is not None:
            container_definition["firelensConfiguration"] = self.firelens_configuration

        return container_definition

    def build_container_environment(self, aws_client: Optional[AwsApiClient] = None) -> List[Dict[str, Any]]:
        logger.debug("Building container environment")
        container_environment: List[Dict[str, Any]] = []
        if self.environment is not None:
            from agno.aws.resource.reference import AwsReference

            for env in self.environment:
                env_name = env.get("name", None)
                env_value = env.get("value", None)
                env_value_parsed = None
                if isinstance(env_value, AwsReference):
                    logger.debug(f"{env_name} is an AwsReference")
                    try:
                        env_value_parsed = env_value.get_reference(aws_client=aws_client)
                    except Exception as e:
                        logger.error(f"Error while parsing {env_name}: {e}")
                else:
                    env_value_parsed = env_value

                if env_value_parsed is not None:
                    try:
                        env_val_str = str(env_value_parsed)
                        container_environment.append({"name": env_name, "value": env_val_str})
                    except Exception as e:
                        logger.error(f"Error while converting {env_value} to str: {e}")

        if self.env_from_secrets is not None:
            secrets: Dict[str, Any] = read_secrets(self.env_from_secrets, aws_client=aws_client)
            for secret_name, secret_value in secrets.items():
                try:
                    secret_value = str(secret_value)
                    container_environment.append({"name": secret_name, "value": secret_value})
                except Exception as e:
                    logger.error(f"Error while converting {secret_value} to str: {e}")
        return container_environment



"""Agno cli

This is the entrypoint for the `agno` cli application.
"""

import typer

from agno.cli.infra_cli import infra_cli as infra_subcommands

agno_cli = typer.Typer(
    help="""\b
Agno is a lightweight framework for building Agent Systems.
\b
Usage:
1. Run `ag infra create` to create a new Agentics Infrastructure project from a template
2. Run `ag infra up` to start the infrastructure
3. Run `ag infra down` to stop the infrastructure
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="[COMMAND] [OPTIONS]",
    pretty_exceptions_show_locals=False,
)


agno_cli.add_typer(infra_subcommands)



from typing import List, Optional

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.utilities.logging import logger


class Subnet(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#subnet
    """

    name: str
    resource_type: Optional[str] = "Subnet"
    service_name: str = "ec2"

    def get_availability_zone(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        # logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        client: AwsApiClient = aws_client or self.get_aws_client()
        service_resource = self.get_service_resource(client)
        try:
            subnet = service_resource.Subnet(self.name)
            az = subnet.availability_zone
            logger.debug(f"AZ for {self.name}: {az}")
            return az
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}: {e}")
        return None

    def get_vpc_id(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        # logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        client: AwsApiClient = aws_client or self.get_aws_client()
        service_resource = self.get_service_resource(client)
        try:
            subnet = service_resource.Subnet(self.name)
            vpc_id = subnet.vpc_id
            logger.debug(f"VPC ID for {self.name}: {vpc_id}")
            return vpc_id
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}: {e}")
        return None


def get_vpc_id_from_subnet_ids(
    subnet_ids: Optional[List[str]], aws_client: Optional[AwsApiClient] = None
) -> Optional[str]:
    if subnet_ids is None:
        return None

    # Get VPC ID from subnets
    vpc_ids = set()
    for subnet in subnet_ids:
        _vpc = Subnet(name=subnet).get_vpc_id(aws_client)
        vpc_ids.add(_vpc)
    if len(vpc_ids) > 1:
        raise ValueError("Subnets must be in the same VPC")
    vpc_id = vpc_ids.pop() if len(vpc_ids) == 1 else None
    return vpc_id



from agno.aws.resource.ec2.security_group import InboundRule, OutboundRule, SecurityGroup, get_my_ip
from agno.aws.resource.ec2.subnet import Subnet
from agno.aws.resource.ec2.volume import EbsVolume

__all__ = [
    "InboundRule",
    "OutboundRule",
    "SecurityGroup",
    "get_my_ip",
    "Subnet",
    "EbsVolume",
]



from pathlib import Path
from typing import Any, Optional

from pydantic import Field

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.utilities.logging import logger


class S3Object(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/object/index.html
    """

    resource_type: str = "s3"
    service_name: str = "s3"

    # The Object's bucket_name identifier. This must be set.
    bucket_name: str
    # The Object's key identifier. This must be set.
    name: str = Field(..., alias="key")

    @property
    def uri(self) -> str:
        """Returns the URI of the s3.Object

        Returns:
            str: The URI of the s3.Object
        """
        return f"s3://{self.bucket_name}/{self.name}"

    def get_resource(self, aws_client: Optional[AwsApiClient] = None) -> Any:
        """Returns the s3.Object

        Args:
            aws_client: The AwsApiClient for the current cluster

        Returns:
            The s3.Object
        """
        client: AwsApiClient = aws_client or self.get_aws_client()
        service_resource = self.get_service_resource(client)
        return service_resource.Object(
            bucket_name=self.bucket_name,
            key=self.name,
        )

    def download(self, path: Path, aws_client: Optional[AwsApiClient] = None) -> None:
        """Downloads the s3.Object to the specified path

        Args:
            path: The path to download the s3.Object to
            aws_client: The AwsApiClient for the current cluster
        """
        logger.info(f"Downloading {self.uri} to {path}")
        object_resource = self.get_resource(aws_client=aws_client)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open(mode="wb") as f:
            object_resource.download_fileobj(f)



from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from agno.base.base import InfraBase
from agno.base.resources import InfraResources
from agno.infra.settings import InfraSettings
from agno.utilities.logging import logger

# List of directories to ignore when loading the Infra
ignored_dirs = ["ignore", "test", "tests", "config"]


class InfraConfig(BaseModel):
    """The InfraConfig holds the configuration for an Agno Infrastructure."""

    # Root directory of the infrastructure.
    infra_root_path: Path

    # Path to the "infra" directory inside the infrastructure root
    internal_infra_dir_path: Optional[Path] = None
    # InfraSettings
    internal_infra_settings: Optional[InfraSettings] = None

    def to_dict(self) -> dict:
        return {
            "infra_root_path": self.infra_root_path,
        }

    @property
    def infra_dir_path(self) -> Optional[Path]:
        if self.internal_infra_dir_path is None:
            if self.infra_root_path is not None:
                from agno.infra.helpers import get_infra_dir_path

                self.internal_infra_dir_path = get_infra_dir_path(self.infra_root_path)
        return self.internal_infra_dir_path

    def validate_infra_settings(self, obj: Any) -> bool:
        if not isinstance(obj, InfraSettings):
            raise Exception("InfraSettings must be of type InfraSettings")

        if self.infra_root_path is not None and obj.infra_root is not None:
            if obj.infra_root != self.infra_root_path:
                raise Exception(f"InfraSettings.infra_root ({obj.infra_root}) must match {self.infra_root_path}")
        return True

    @property
    def infra_settings(self) -> Optional[InfraSettings]:
        if self.internal_infra_settings is not None:
            return self.internal_infra_settings

        infra_settings_file: Optional[Path] = None
        if self.infra_dir_path is not None:
            _infra_settings_file_path = self.infra_dir_path.joinpath("settings.py")
            if _infra_settings_file_path.exists() and _infra_settings_file_path.is_file():
                infra_settings_file = _infra_settings_file_path
        if infra_settings_file is None:
            logger.debug("infra_settings file not found")
            return None

        logger.debug(f"Loading infra_settings from {infra_settings_file}")
        try:
            from agno.utilities.py_io import get_python_objects_from_module

            python_objects = get_python_objects_from_module(infra_settings_file)
            for obj_name, obj in python_objects.items():
                if isinstance(obj, InfraSettings):
                    if self.validate_infra_settings(obj):
                        self.internal_infra_settings = obj
        except Exception:
            logger.warning(f"Error in {infra_settings_file}")
            raise
        return self.internal_infra_settings

    def set_local_env(self) -> None:
        from os import environ

        from agno.constants import (
            AGNO_INFRA_DIR,
            AGNO_INFRA_NAME,
            AGNO_INFRA_ROOT,
            AWS_REGION_ENV_VAR,
        )

        if self.infra_root_path is not None:
            environ[AGNO_INFRA_ROOT] = str(self.infra_root_path)

            infra_dir_path: Optional[Path] = self.infra_dir_path
            if infra_dir_path is not None:
                environ[AGNO_INFRA_DIR] = str(infra_dir_path)

            if self.infra_settings is not None:
                environ[AGNO_INFRA_NAME] = str(self.infra_settings.infra_name)

        if (
            environ.get(AWS_REGION_ENV_VAR) is None
            and self.infra_settings is not None
            and self.infra_settings.aws_region is not None
        ):
            environ[AWS_REGION_ENV_VAR] = self.infra_settings.aws_region

    def get_resources(
        self,
        env: Optional[str] = None,
        infra: Optional[str] = None,
        order: str = "create",
    ) -> List[InfraResources]:
        if self.infra_root_path is None:
            logger.warning("InfraConfig.infra_root_path is None")
            return []

        from sys import path as sys_path

        from agno.utilities.load_env import load_env
        from agno.utilities.py_io import get_python_objects_from_module

        logger.debug("**--> Loading InfraConfig")
        logger.debug(f"Loading .env from {self.infra_root_path}")
        load_env(dotenv_dir=self.infra_root_path)

        # NOTE: When loading a Infra, relative imports or package imports do not work.
        # This is a known problem in python
        #     eg: https://stackoverflow.com/questions/6323860/sibling-package-imports/50193944#50193944
        # To make them work, we add infra_root to sys.path so is treated as a module
        logger.debug(f"Adding {self.infra_root_path} to path")
        sys_path.insert(0, str(self.infra_root_path))

        infra_objects: Dict[str, InfraResources] = {}

        infra_dir_path: Optional[Path] = self.infra_dir_path
        if infra_dir_path is not None:
            logger.debug(f"--^^-- Loading Infra from: {infra_dir_path}")
            # Create a dict of objects in the Infra directory
            resource_files = infra_dir_path.rglob("*.py")
            for resource_file in resource_files:
                if resource_file.name == "__init__.py":
                    continue

                resource_file_parts = resource_file.parts
                infra_dir_path_parts = infra_dir_path.parts
                resource_file_parts_after_ws = resource_file_parts[len(infra_dir_path_parts) :]
                # Check if file in ignored directory
                if any([ignored_dir in resource_file_parts_after_ws for ignored_dir in ignored_dirs]):
                    logger.debug(f"Skipping file in ignored directory: {resource_file}")
                    continue
                logger.debug(f"Reading file: {resource_file}")
                try:
                    python_objects = get_python_objects_from_module(resource_file)
                    # logger.debug(f"python_objects: {python_objects}")
                    for obj_name, obj in python_objects.items():
                        if isinstance(obj, InfraSettings):
                            logger.debug(f"Found: {obj.__class__.__module__}: {obj_name}")
                            if self.validate_infra_settings(obj):
                                self.internal_infra_settings = obj
                        elif isinstance(obj, InfraResources):
                            logger.debug(f"Found: {obj.__class__.__module__}: {obj_name}")
                            if not obj.enabled:
                                logger.debug(f"Skipping {obj_name}: disabled")
                                continue
                            infra_objects[obj_name] = obj
                except Exception:
                    logger.warning(f"Error in {resource_file}")
                    raise
            logger.debug(f"infra_objects: {infra_objects}")
        logger.debug("**--> InfraConfig loaded")
        logger.debug(f"Removing {self.infra_root_path} from path")
        sys_path.remove(str(self.infra_root_path))

        # Filter resources by infra
        filtered_infra_objects_by_infra_type: Dict[str, InfraResources] = {}
        logger.debug(f"Filtering resources for env: {env} | infra: {infra} | order: {order}")
        if infra is None:
            filtered_infra_objects_by_infra_type = infra_objects
        else:
            for resource_name, resource in infra_objects.items():
                if resource.infra == infra:
                    filtered_infra_objects_by_infra_type[resource_name] = resource

        # Filter resources by env
        filtered_infra_objects_by_env: Dict[str, InfraResources] = {}
        if env is None:
            filtered_infra_objects_by_env = filtered_infra_objects_by_infra_type
        else:
            for resource_name, resource in filtered_infra_objects_by_infra_type.items():
                if resource.env == env:
                    filtered_infra_objects_by_env[resource_name] = resource

        # Updated resources with the infra settings
        # Create a temporary infra settings object if it does not exist
        if self.infra_settings is None:
            self.internal_infra_settings = InfraSettings(
                infra_root=self.infra_root_path,
                infra_name=self.infra_root_path.stem,
            )
            logger.debug(f"Created InfraSettings: {self.infra_settings}")
        # Update the resources with the infra settings
        if self.infra_settings is not None:
            for resource_name, resource in filtered_infra_objects_by_env.items():
                logger.debug(f"Setting infra settings for {resource.__class__.__name__}")
                resource.set_infra_settings(self.infra_settings)

        # Create a list of InfraResources from the filtered resources
        infra_resources_list: List[InfraResources] = []
        for resource_name, resource in filtered_infra_objects_by_env.items():
            # If the resource is an InfraResources object, add it to the list
            if isinstance(resource, InfraResources):
                infra_resources_list.append(resource)

        return infra_resources_list

    @staticmethod
    def get_resources_from_file(
        resource_file: Path,
        env: Optional[str] = None,
        infra: Optional[str] = None,
        order: str = "create",
    ) -> List[InfraResources]:
        if not resource_file.exists():
            raise FileNotFoundError(f"File {resource_file} does not exist")
        if not resource_file.is_file():
            raise ValueError(f"Path {resource_file} is not a file")
        if not resource_file.suffix == ".py":
            raise ValueError(f"File {resource_file} is not a python file")

        from sys import path as sys_path

        from agno.utilities.load_env import load_env
        from agno.utilities.py_io import get_python_objects_from_module

        resource_file_parent_dir = resource_file.parent.resolve()
        logger.debug(f"Loading .env from {resource_file_parent_dir}")
        load_env(dotenv_dir=resource_file_parent_dir)

        temporary_infra_config = InfraConfig(infra_root_path=resource_file_parent_dir)

        # NOTE: When loading a directory, relative imports or package imports do not work.
        # This is a known problem in python
        #     eg: https://stackoverflow.com/questions/6323860/sibling-package-imports/50193944#50193944
        # To make them work, we add the resource_file_parent_dir to sys.path so it can be treated as a module
        logger.debug(f"Adding {resource_file_parent_dir} to path")
        sys_path.insert(0, str(resource_file_parent_dir))

        logger.debug(f"**--> Reading Infra resources from {resource_file}")

        # Get all infra resources from the file
        infra_objects: Dict[str, InfraBase] = {}
        try:
            # Get all python objects from the file
            python_objects = get_python_objects_from_module(resource_file)
            # Filter out the objects that are subclasses of InfraSettings
            for obj_name, obj in python_objects.items():
                if isinstance(obj, InfraSettings):
                    logger.debug(f"Found: {obj.__class__.__module__}: {obj_name}")
                    infra_objects[obj_name] = obj  # type: ignore
        except Exception:
            logger.error(f"Error reading: {resource_file}")
            raise

        # Filter resources by infra
        filtered_infra_objects_by_infra_type: Dict[str, InfraBase] = {}
        logger.debug(f"Filtering resources for env: {env} | infra: {infra} | order: {order}")
        if infra is None:
            filtered_infra_objects_by_infra_type = infra_objects
        else:
            for resource_name, resource in infra_objects.items():
                if resource.infra == infra:
                    filtered_infra_objects_by_infra_type[resource_name] = resource

        # Filter resources by env
        filtered_infra_objects_by_env: Dict[str, InfraBase] = {}
        if env is None:
            filtered_infra_objects_by_env = filtered_infra_objects_by_infra_type
        else:
            for resource_name, resource in filtered_infra_objects_by_infra_type.items():
                if resource.env == env:
                    filtered_infra_objects_by_env[resource_name] = resource

        # Updated resources with the infra settings
        # Create a temporary infra settings object if it does not exist
        if temporary_infra_config.infra_settings is None:
            temporary_infra_config.internal_infra_settings = InfraSettings(
                infra_root=temporary_infra_config.infra_root_path,
                infra_name=temporary_infra_config.infra_root_path.stem,
            )
        # Update the resources with the infra settings
        if temporary_infra_config.infra_settings is not None:
            for resource_name, resource in filtered_infra_objects_by_env.items():
                logger.debug(f"Setting infra settings for {resource.__class__.__name__}")
                resource.set_infra_settings(temporary_infra_config.infra_settings)

        # Create a list of InfraResources from the filtered resources
        infra_resources_list: List[InfraResources] = []
        for resource_name, resource in filtered_infra_objects_by_env.items():
            # If the resource is an InfraResources object, add it to the list
            if isinstance(resource, InfraResources):
                infra_resources_list.append(resource)
            # Otherwise, get the InfraResources object from the resource
            else:
                _infra_resources = resource.get_infra_resources()
                if _infra_resources is not None and isinstance(_infra_resources, InfraResources):
                    infra_resources_list.append(_infra_resources)

        return infra_resources_list



from agno.aws.resource.ecs.cluster import EcsCluster
from agno.aws.resource.ecs.container import EcsContainer
from agno.aws.resource.ecs.service import EcsService
from agno.aws.resource.ecs.task_definition import EcsTaskDefinition
from agno.aws.resource.ecs.volume import EcsVolume



from collections import OrderedDict
from typing import Dict, List, Type, Union

from agno.aws.resource.acm.certificate import AcmCertificate
from agno.aws.resource.base import AwsResource
from agno.aws.resource.cloudformation.stack import CloudFormationStack
from agno.aws.resource.ec2.security_group import SecurityGroup
from agno.aws.resource.ec2.subnet import Subnet
from agno.aws.resource.ec2.volume import EbsVolume
from agno.aws.resource.ecs.cluster import EcsCluster
from agno.aws.resource.ecs.service import EcsService
from agno.aws.resource.ecs.task_definition import EcsTaskDefinition
from agno.aws.resource.elasticache.cluster import CacheCluster
from agno.aws.resource.elasticache.subnet_group import CacheSubnetGroup
from agno.aws.resource.elb.listener import Listener
from agno.aws.resource.elb.load_balancer import LoadBalancer
from agno.aws.resource.elb.target_group import TargetGroup
from agno.aws.resource.emr.cluster import EmrCluster
from agno.aws.resource.glue.crawler import GlueCrawler
from agno.aws.resource.iam.policy import IamPolicy
from agno.aws.resource.iam.role import IamRole
from agno.aws.resource.rds.db_cluster import DbCluster
from agno.aws.resource.rds.db_instance import DbInstance
from agno.aws.resource.rds.db_subnet_group import DbSubnetGroup
from agno.aws.resource.s3.bucket import S3Bucket
from agno.aws.resource.secret.manager import SecretsManager

# Use this as a type for an object which can hold any AwsResource
AwsResourceType = Union[
    AcmCertificate,
    CloudFormationStack,
    EbsVolume,
    IamRole,
    IamPolicy,
    GlueCrawler,
    S3Bucket,
    SecretsManager,
    Subnet,
    SecurityGroup,
    DbSubnetGroup,
    DbCluster,
    DbInstance,
    CacheSubnetGroup,
    CacheCluster,
    EmrCluster,
    EcsCluster,
    EcsTaskDefinition,
    EcsService,
    LoadBalancer,
    TargetGroup,
    Listener,
]

# Use this as an ordered list to iterate over all AwsResource Classes
# This list is the order in which resources should be installed as well.
AwsResourceTypeList: List[Type[AwsResource]] = [
    Subnet,
    SecurityGroup,
    IamRole,
    IamPolicy,
    S3Bucket,
    SecretsManager,
    EbsVolume,
    AcmCertificate,
    CloudFormationStack,
    GlueCrawler,
    DbSubnetGroup,
    DbCluster,
    DbInstance,
    CacheSubnetGroup,
    CacheCluster,
    LoadBalancer,
    TargetGroup,
    Listener,
    EcsCluster,
    EcsTaskDefinition,
    EcsService,
    EmrCluster,
]

# Map Aws resource alias' to their type
_aws_resource_type_names: Dict[str, Type[AwsResource]] = {
    aws_type.__name__.lower(): aws_type for aws_type in AwsResourceTypeList
}
_aws_resource_type_aliases: Dict[str, Type[AwsResource]] = {
    "s3": S3Bucket,
    "volume": EbsVolume,
}

AwsResourceAliasToTypeMap: Dict[str, Type[AwsResource]] = dict(**_aws_resource_type_names, **_aws_resource_type_aliases)

# Maps each AwsResource to an install weight
# lower weight AwsResource(s) get installed first
AwsResourceInstallOrder: Dict[str, int] = OrderedDict(
    {resource_type.__name__: idx for idx, resource_type in enumerate(AwsResourceTypeList, start=1)}
)



from collections import OrderedDict
from typing import Dict, List, Type, Union

from agno.docker.resource.base import DockerResource
from agno.docker.resource.container import DockerContainer
from agno.docker.resource.image import DockerImage
from agno.docker.resource.network import DockerNetwork
from agno.docker.resource.volume import DockerVolume

# Use this as a type for an object that can hold any DockerResource
DockerResourceType = Union[
    DockerNetwork,
    DockerImage,
    DockerVolume,
    DockerContainer,
]

# Use this as an ordered list to iterate over all DockerResource Classes
# This list is the order in which resources are installed as well.
DockerResourceTypeList: List[Type[DockerResource]] = [
    DockerNetwork,
    DockerImage,
    DockerVolume,
    DockerContainer,
]

# Maps each DockerResource to an Install weight
# lower weight DockerResource(s) get installed first
# i.e. Networks are installed first, Images, then Volumes ... and so on
DockerResourceInstallOrder: Dict[str, int] = OrderedDict(
    {resource_type.__name__: idx for idx, resource_type in enumerate(DockerResourceTypeList, start=1)}
)



from typing import Optional

from agno.base.app import InfraApp


class DbApp(InfraApp):
    db_user: Optional[str] = None
    db_password: Optional[str] = None
    db_database: Optional[str] = None
    db_driver: Optional[str] = None

    def get_db_user(self) -> Optional[str]:
        return self.db_user or self.get_secret_from_file("DB_USER")

    def get_db_password(self) -> Optional[str]:
        return self.db_password or self.get_secret_from_file("DB_PASSWORD")

    def get_db_database(self) -> Optional[str]:
        return self.db_database or self.get_secret_from_file("DB_DATABASE")

    def get_db_driver(self) -> Optional[str]:
        return self.db_driver or self.get_secret_from_file("DB_DRIVER")

    def get_db_host(self) -> Optional[str]:
        raise NotImplementedError

    def get_db_port(self) -> Optional[int]:
        raise NotImplementedError

    def get_db_connection(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        database = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host()
        port = self.get_db_port()
        return f"{driver}://{user}:{password}@{host}:{port}/{database}"

    def get_db_host_local(self) -> Optional[str]:
        return "localhost"

    def get_db_port_local(self) -> Optional[int]:
        return self.host_port

    def get_db_connection_local(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        database = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{user}:{password}@{host}:{port}/{database}"



from typing import Any, Dict, List, Optional, Union

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.aws.resource.ec2.security_group import SecurityGroup
from agno.aws.resource.ec2.subnet import Subnet
from agno.cli.console import print_info
from agno.utilities.logging import logger


class LoadBalancer(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/elbv2.html
    """

    resource_type: Optional[str] = "LoadBalancer"
    service_name: str = "elbv2"

    # Name of the Load Balancer.
    name: str
    subnets: Optional[List[Union[str, Subnet]]] = None
    subnet_mappings: Optional[List[Dict[str, str]]] = None
    security_groups: Optional[List[Union[str, SecurityGroup]]] = None
    scheme: Optional[str] = None
    tags: Optional[List[Dict[str, str]]] = None
    type: Optional[str] = None
    ip_address_type: Optional[str] = None
    customer_owned_ipv_4_pool: Optional[str] = None

    # Protocol for load_balancer: HTTP or HTTPS
    protocol: str = "HTTP"

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the Load Balancer

        Args:
            aws_client: The AwsApiClient for the current Load Balancer
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}

        if self.subnets is not None:
            subnet_ids = []
            for subnet in self.subnets:
                if isinstance(subnet, Subnet):
                    subnet_ids.append(subnet.name)
                elif isinstance(subnet, str):
                    subnet_ids.append(subnet)
            not_null_args["Subnets"] = subnet_ids

        if self.subnet_mappings is not None:
            not_null_args["SubnetMappings"] = self.subnet_mappings

        if self.security_groups is not None:
            security_group_ids = []
            for sg in self.security_groups:
                if isinstance(sg, SecurityGroup):
                    security_group_ids.append(sg.get_security_group_id(aws_client))
                else:
                    security_group_ids.append(sg)
            not_null_args["SecurityGroups"] = security_group_ids

        if self.scheme is not None:
            not_null_args["Scheme"] = self.scheme
        if self.tags is not None:
            not_null_args["tags"] = self.tags
        if self.type is not None:
            not_null_args["Type"] = self.type
        if self.ip_address_type is not None:
            not_null_args["IpAddressType"] = self.ip_address_type
        if self.customer_owned_ipv_4_pool is not None:
            not_null_args["CustomerOwnedIpv4Pool"] = self.customer_owned_ipv_4_pool

        # Create LoadBalancer
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_load_balancer(
                Name=self.name,
                **not_null_args,
            )
            logger.debug(f"Create Response: {create_response}")
            resource_dict = create_response.get("LoadBalancers", {})

            # Validate resource creation
            if resource_dict is not None:
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for LoadBalancer to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                waiter = self.get_service_client(aws_client).get_waiter("load_balancer_exists")
                waiter.wait(
                    Names=[self.get_resource_name()],
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        # Read the LoadBalancer
        elb = self._read(aws_client)
        if elb is None:
            logger.error(f"Error reading {self.get_resource_type()}. Please get DNS name manually.")
        else:
            dns_name = elb.get("DNSName", None)
            print_info(f"LoadBalancer DNS: {self.protocol.lower()}://{dns_name}")
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the LoadBalancer

        Args:
            aws_client: The AwsApiClient for the current LoadBalancer
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_load_balancers(Names=[self.name])
            logger.debug(f"Describe Response: {describe_response}")
            resource_list = describe_response.get("LoadBalancers", None)

            if resource_list is not None and isinstance(resource_list, list):
                self.active_resource = resource_list[0]
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the LoadBalancer

        Args:
            aws_client: The AwsApiClient for the current LoadBalancer
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        try:
            lb_arn = self.get_arn(aws_client)
            if lb_arn is None:
                logger.warning(f"{self.get_resource_type()} not found.")
                return True
            delete_response = service_client.delete_load_balancer(LoadBalancerArn=lb_arn)
            logger.debug(f"Delete Response: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for LoadBalancer to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("load_balancers_deleted")
                waiter.wait(
                    Names=[self.get_resource_name()],
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        lb = self._read(aws_client)
        if lb is None:
            return None
        lb_arn = lb.get("LoadBalancerArn", None)
        return lb_arn



PYTHONPATH_ENV_VAR: str = "PYTHONPATH"
AGNO_RUNTIME_ENV_VAR: str = "AGNO_RUNTIME"

AGNO_INFRA_NAME: str = "AGNO_INFRA_NAME"
AGNO_INFRA_ROOT: str = "AGNO_INFRA_ROOT"
AGNO_INFRA_DIR: str = "AGNO_INFRA_DIR"
REQUIREMENTS_FILE_PATH_ENV_VAR: str = "REQUIREMENTS_FILE_PATH"

AWS_REGION_ENV_VAR: str = "AWS_REGION"
AWS_DEFAULT_REGION_ENV_VAR: str = "AWS_DEFAULT_REGION"
AWS_PROFILE_ENV_VAR: str = "AWS_PROFILE"



from pathlib import Path
from typing import Optional

import git

from agno.utilities.logging import log_debug


def get_remote_origin_for_dir(
    ws_root_path: Optional[Path],
) -> Optional[str]:
    """Returns the remote origin for a directory"""

    if ws_root_path is None or not ws_root_path.exists() or not ws_root_path.is_dir():
        return None

    _remote_origin: Optional[git.Remote] = None
    try:
        _git_repo: git.Repo = git.Repo(path=ws_root_path)
        _remote_origin = _git_repo.remote("origin")
    except (git.InvalidGitRepositoryError, ValueError):
        return None
    except git.NoSuchPathError:
        return None

    if _remote_origin is None:
        return None

    # TODO: Figure out multiple urls for origin and how to only get the fetch url
    # _remote_origin.urls returns a generator
    _remote_origin_url: Optional[str] = None
    for _remote_url in _remote_origin.urls:
        _remote_origin_url = _remote_url
        break
    return _remote_origin_url


class GitCloneProgress(git.RemoteProgress):
    # https://gitpython.readthedocs.io/en/stable/reference.html#module-git.remote
    # def line_dropped(self, line):
    #     print("line dropped: {}".format(line))

    def update(self, op_code, cur_count, max_count=None, message=""):
        if op_code == 5:
            log_debug("Starting copy")
        if op_code == 10:
            log_debug("Copy complete")
        # log_debug(f"op_code: {op_code}")
        # log_debug(f"cur_count: {cur_count}")
        # log_debug(f"max_count: {max_count}")
        # log_debug(f"message: {message}")
        # print(self._cur_line)



from typing import Dict, Optional

from agno.base.db_app import DbApp
from agno.docker.app.base import ContainerContext, DockerApp  # noqa: F401


class PostgresDb(DockerApp, DbApp):
    # -*- App Name
    name: str = "postgres"

    # -*- Image Configuration
    image_name: str = "postgres"
    image_tag: str = "17.2"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = True
    port_number: int = 5432

    # -*- Postgres Volume
    # Create a volume for postgres storage
    create_volume: bool = True
    # Path to mount the volume inside the container
    volume_container_path: str = "/var/lib/postgresql/data"

    # -*- Postgres Configuration
    # Provide POSTGRES_USER as pg_user or POSTGRES_USER in secrets_file
    pg_user: Optional[str] = None
    # Provide POSTGRES_PASSWORD as pg_password or POSTGRES_PASSWORD in secrets_file
    pg_password: Optional[str] = None
    # Provide POSTGRES_DB as pg_database or POSTGRES_DB in secrets_file
    pg_database: Optional[str] = None
    pg_driver: str = "postgresql+psycopg"
    pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata"
    postgres_initdb_args: Optional[str] = None
    postgres_initdb_waldir: Optional[str] = None
    postgres_host_auth_method: Optional[str] = None
    postgres_password_file: Optional[str] = None
    postgres_user_file: Optional[str] = None
    postgres_db_file: Optional[str] = None
    postgres_initdb_args_file: Optional[str] = None

    def get_db_user(self) -> Optional[str]:
        return self.pg_user or self.get_secret_from_file("POSTGRES_USER")

    def get_db_password(self) -> Optional[str]:
        return self.pg_password or self.get_secret_from_file("POSTGRES_PASSWORD")

    def get_db_database(self) -> Optional[str]:
        return self.pg_database or self.get_secret_from_file("POSTGRES_DB")

    def get_db_driver(self) -> Optional[str]:
        return self.pg_driver

    def get_db_host(self) -> Optional[str]:
        return self.get_container_name()

    def get_db_port(self) -> Optional[int]:
        return self.container_port

    def get_container_env(self, container_context: ContainerContext) -> Dict[str, str]:
        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}

        # Set postgres env vars
        # Check: https://hub.docker.com/_/postgres
        db_user = self.get_db_user()
        if db_user:
            container_env["POSTGRES_USER"] = db_user
        db_password = self.get_db_password()
        if db_password:
            container_env["POSTGRES_PASSWORD"] = db_password
        db_database = self.get_db_database()
        if db_database:
            container_env["POSTGRES_DB"] = db_database
        if self.pgdata:
            container_env["PGDATA"] = self.pgdata
        if self.postgres_initdb_args:
            container_env["POSTGRES_INITDB_ARGS"] = self.postgres_initdb_args
        if self.postgres_initdb_waldir:
            container_env["POSTGRES_INITDB_WALDIR"] = self.postgres_initdb_waldir
        if self.postgres_host_auth_method:
            container_env["POSTGRES_HOST_AUTH_METHOD"] = self.postgres_host_auth_method
        if self.postgres_password_file:
            container_env["POSTGRES_PASSWORD_FILE"] = self.postgres_password_file
        if self.postgres_user_file:
            container_env["POSTGRES_USER_FILE"] = self.postgres_user_file
        if self.postgres_db_file:
            container_env["POSTGRES_DB_FILE"] = self.postgres_db_file
        if self.postgres_initdb_args_file:
            container_env["POSTGRES_INITDB_ARGS_FILE"] = self.postgres_initdb_args_file

        # Update the container env using env_file
        env_data_from_file = self.get_env_file_data()
        if env_data_from_file is not None:
            container_env.update({k: str(v) for k, v in env_data_from_file.items() if v is not None})

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_file_data()
        if secret_data_from_file is not None:
            container_env.update({k: str(v) for k, v in secret_data_from_file.items() if v is not None})

        # Update the container env with user provided env_vars
        # this overwrites any existing variables with the same key
        if self.env_vars is not None and isinstance(self.env_vars, dict):
            container_env.update({k: str(v) for k, v in self.env_vars.items() if v is not None})

        return container_env



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional

from agno.cli.console import print_heading, print_info
from agno.cli.settings import agno_cli_settings
from agno.infra.config import InfraConfig
from agno.utilities.json_io import read_json_file, write_json_file
from agno.utilities.logging import logger


class AgnoCliConfig:
    """The AgnoCliConfig class manages user data for the agno cli"""

    def __init__(
        self,
        active_infra_dir: Optional[str] = None,
        infra_config_map: Optional[Dict[str, InfraConfig]] = None,
    ) -> None:
        # Active infra dir - used as the default for `ag` commands
        # To add an active infra, use the active_infra_dir setter
        self._active_infra_dir: Optional[str] = active_infra_dir

        # Mapping from infra_root_path to infra_config
        self.infra_config_map: Dict[str, InfraConfig] = infra_config_map or OrderedDict()

    ######################################################
    ## Infra functions
    ######################################################

    @property
    def active_infra_dir(self) -> Optional[str]:
        return self._active_infra_dir

    def set_active_infra_dir(self, infra_root_path: Optional[Path]) -> None:
        if infra_root_path is not None:
            logger.debug(f"Setting active infra to: {str(infra_root_path)}")
            self._active_infra_dir = str(infra_root_path)
            self.save_config()

    @property
    def available_infra(self) -> List[InfraConfig]:
        return list(self.infra_config_map.values())

    def _add_or_update_infra_config(
        self,
        infra_root_path: Path,
    ) -> Optional[InfraConfig]:
        """The main function to create, update or refresh a InfraConfig.

        This function does not call self.save_config(). Remember to save_config() after calling this function.
        """

        # Validate infra_root_path
        if infra_root_path is None or not isinstance(infra_root_path, Path):
            raise ValueError(f"Invalid infra_root: {infra_root_path}")
        infra_root_str = str(infra_root_path)

        ######################################################
        # Create new infra_config if one does not exist
        ######################################################
        if infra_root_str not in self.infra_config_map:
            logger.debug(f"Creating infra at: {infra_root_str}")
            new_infra_config = InfraConfig(
                infra_root_path=infra_root_path,
            )
            self.infra_config_map[infra_root_str] = new_infra_config
            logger.debug(f"Infra created at: {infra_root_str}")

            # Return the new_infra_config
            return new_infra_config

        ######################################################
        # Update infra_config
        ######################################################
        logger.debug(f"Updating infra at: {infra_root_str}")
        # By this point there should be a InfraConfig object for this infra_name
        existing_infra_config: Optional[InfraConfig] = self.infra_config_map.get(infra_root_str, None)
        if existing_infra_config is None:
            logger.error(f"Could not find infra at: {infra_root_str}, please run `ag infra setup`")
            return None

        # Swap the existing infra_config with the updated one
        self.infra_config_map[infra_root_str] = existing_infra_config

        # Return the updated_infra_config
        return existing_infra_config

    def add_new_infra_to_config(self, infra_root_path: Path) -> Optional[InfraConfig]:
        """Adds a newly created workspace to the AgnoCliConfig"""

        infra_config = self._add_or_update_infra_config(infra_root_path=infra_root_path)
        self.save_config()
        return infra_config

    def create_or_update_infra_config(
        self,
        infra_root_path: Path,
        set_as_active: bool = True,
    ) -> Optional[InfraConfig]:
        """Creates or updates a WorkspaceConfig and returns the WorkspaceConfig"""

        infra_config = self._add_or_update_infra_config(infra_root_path=infra_root_path)
        if set_as_active:
            self._active_infra_dir = str(infra_root_path)
        self.save_config()
        return infra_config

    def delete_infra(self, infra_root_path: Path) -> None:
        """Handles Deleting a infra from the AgnoCliConfig and api"""

        infra_root_str = str(infra_root_path)
        print_heading(f"Deleting record for infra: {infra_root_str}")

        infra_config: Optional[InfraConfig] = self.infra_config_map.pop(infra_root_str, None)
        if infra_config is None:
            logger.warning(f"No record of infra at {infra_root_str}")
            return

        # Check if we're deleting the active infra, if yes, unset the active infra
        if self._active_infra_dir is not None and self._active_infra_dir == infra_root_str:
            print_info(f"Removing {infra_root_str} as the active infra")
            self._active_infra_dir = None
        self.save_config()
        print_info("Infra record deleted")

    def get_infra_config_by_dir_name(self, infra_dir_name: str) -> Optional[InfraConfig]:
        infra_root_str: Optional[str] = None
        for k, v in self.infra_config_map.items():
            if v.infra_root_path.stem == infra_dir_name:
                infra_root_str = k
                break

        if infra_root_str is None or infra_root_str not in self.infra_config_map:
            return None

        return self.infra_config_map[infra_root_str]

    def get_infra_config_by_path(self, infra_root_path: Path) -> Optional[InfraConfig]:
        return self.infra_config_map[str(infra_root_path)] if str(infra_root_path) in self.infra_config_map else None

    def get_active_infra_config(self) -> Optional[InfraConfig]:
        if self.active_infra_dir is not None and self.active_infra_dir in self.infra_config_map:
            return self.infra_config_map[self.active_infra_dir]
        return None

    ######################################################
    ## Save AgnoCliConfig
    ######################################################

    def save_config(self):
        config_data = {
            "active_infra_dir": self.active_infra_dir,
            "infra_config_map": {k: v.to_dict() for k, v in self.infra_config_map.items()},
        }
        write_json_file(file_path=agno_cli_settings.config_file_path, data=config_data)

    @classmethod
    def from_saved_config(cls) -> Optional["AgnoCliConfig"]:
        try:
            config_data = read_json_file(file_path=agno_cli_settings.config_file_path)
            if config_data is None or not isinstance(config_data, dict):
                logger.debug("No config found")
                return None

            active_infra_dir = config_data.get("active_infra_dir")

            # Create a new config
            new_config = cls(active_infra_dir)

            # Add all the workspaces
            for k, v in config_data.get("infra_config_map", {}).items():
                _infra_config = InfraConfig(**v)
                if _infra_config is not None:
                    new_config.infra_config_map[k] = _infra_config
            return new_config
        except Exception as e:
            logger.warning(e)
            logger.warning("Please setup the infra using `ag infra setup`")
            return None

    ######################################################
    ## Print AgnoCliConfig
    ######################################################

    def print_to_cli(self, show_all: bool = False):
        if self.active_infra_dir:
            print_heading(f"Active infra directory: {self.active_infra_dir}\n")
        else:
            print_info("No active infra found.")
            print_info(
                "Please create a infra using `ag infra create` or setup an existing infra using `ag infra setup`"
            )

        if show_all and len(self.infra_config_map) > 0:
            print_heading("Available infra:\n")
            c = 1
            for k, _ in self.infra_config_map.items():
                print_info(f"  {c}. Path: {k}")
                c += 1



from pathlib import Path
from typing import Dict, List, Optional, cast

from rich.prompt import Prompt

from agno.base.resources import InfraResources
from agno.cli.config import AgnoCliConfig
from agno.cli.console import (
    console,
    log_config_not_available_msg,
    print_heading,
    print_info,
    print_subheading,
)
from agno.infra.config import InfraConfig
from agno.infra.enums import InfraStarterTemplate
from agno.utilities.logging import log_warning, logger
from agno.utils.log import log_info

TEMPLATE_TO_NAME_MAP: Dict[InfraStarterTemplate, str] = {
    InfraStarterTemplate.agentos_docker: "agentos-docker",
    InfraStarterTemplate.agentos_aws: "agentos-aws",
    InfraStarterTemplate.agentos_railway: "agentos-railway",
}
TEMPLATE_TO_REPO_MAP: Dict[InfraStarterTemplate, str] = {
    InfraStarterTemplate.agentos_docker: "https://github.com/agno-agi/agentos-docker-template",
    InfraStarterTemplate.agentos_aws: "https://github.com/agno-agi/agentos-aws-template",
    InfraStarterTemplate.agentos_railway: "https://github.com/agno-agi/agentos-railway-template",
}


def create_infra_from_template(
    name: Optional[str] = None,
    template: Optional[str] = None,
    url: Optional[str] = None,
) -> Optional[InfraConfig]:
    """Creates a new AgentOS infra codebase from a template and returns the InfraConfig.

    This function clones a template or url on the users machine at the path:
        cwd/name
    """
    from shutil import copytree

    import git

    from agno.cli.operator import initialize_agno_cli
    from agno.infra.helpers import get_infra_dir_path, is_docker_only_project
    from agno.utilities.filesystem import rmdir_recursive
    from agno.utilities.git import GitCloneProgress

    current_dir: Path = Path("").resolve()

    # Initialize Agno before creating a Infra
    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return None
    agno_config = cast(AgnoCliConfig, agno_config)

    infra_dir_name: Optional[str] = name
    repo_to_clone: Optional[str] = url
    infra_template = InfraStarterTemplate.agentos_docker
    templates = list(InfraStarterTemplate.__members__.values())

    print_subheading("Creating a new AgentOS codebase...\n")

    if repo_to_clone is None:
        # Get repo_to_clone from template
        if template is None:
            # Get starter template from the user if template is not provided
            # Display available starter templates and ask user to select one
            print_info("Select starter template or press Enter for default (agentos-docker)")
            for template_id, template_name in enumerate(templates, start=1):
                print_info("  [b][{}][/b] {}".format(template_id, InfraStarterTemplate(template_name).value))

            # Get starter template from the user
            template_choices = [str(idx) for idx, _ in enumerate(templates, start=1)]
            template_inp_raw = Prompt.ask(
                "Chosen Template",
                choices=template_choices,
                default="1",
                show_choices=False,
            )
            # Convert input to int
            template_inp = int(template_inp_raw) if template_inp_raw is not None else None

            if template_inp is not None:
                template_inp_idx = template_inp - 1
                infra_template = InfraStarterTemplate(templates[template_inp_idx])
        elif template.lower() in InfraStarterTemplate.__members__.values():
            infra_template = InfraStarterTemplate(template)
        else:
            raise Exception(f"{template} is not a supported template, please choose from: {templates}")

        logger.debug(f"Selected Template: {infra_template.value}")
        repo_to_clone = TEMPLATE_TO_REPO_MAP.get(infra_template)

    if infra_dir_name is None:
        default_infra_name = "agno-infra-project"
        if url is not None:
            # Get default_infra_name from url
            default_infra_name = url.split("/")[-1].split(".")[0]
        else:
            # Get default_infra_name from template
            default_infra_name = TEMPLATE_TO_NAME_MAP.get(infra_template, "agentos-docker")
        logger.debug(f"Asking for infra name with default: {default_infra_name}")
        # Ask user for infra name if not provided
        infra_dir_name = Prompt.ask("Infra Project Directory", default=default_infra_name, console=console)

    if repo_to_clone is None:
        logger.error("URL or Template is required")
        return None

    # Check if we can create the infra in the current dir
    infra_root_path: Path = current_dir.joinpath(infra_dir_name)
    if infra_root_path.exists():
        logger.error(
            f"Directory {infra_root_path} exists, please delete the directory or choose another name for your Agno Infra project."
        )
        return None

    print_info("\nCreating your new AgentOS codebase...")
    logger.debug("Cloning: {}".format(repo_to_clone))
    try:
        git.Repo.clone_from(
            repo_to_clone,
            str(infra_root_path),
            progress=GitCloneProgress(),  # type: ignore
        )
    except Exception as e:
        logger.error(e)
        return None

    # Remove existing .git folder
    _dot_git_folder = infra_root_path.joinpath(".git")
    _dot_git_exists = _dot_git_folder.exists()
    if _dot_git_exists:
        logger.debug(f"Deleting {_dot_git_folder}")
        try:
            _dot_git_exists = not rmdir_recursive(_dot_git_folder)
        except Exception as e:
            log_warning(f"Failed to delete {_dot_git_folder}: {e}")
            log_info("Please delete the .git folder manually")
            pass

    agno_config.add_new_infra_to_config(infra_root_path=infra_root_path)

    if is_docker_only_project(infra_root_path):
        logger.debug("Docker only project detected - skipping secrets setup")
    else:
        try:
            # infra_dir_path is the path to the infra_root/infra dir
            infra_dir_path: Optional[Path] = get_infra_dir_path(infra_root_path)
            if infra_dir_path is not None:
                infra_secrets_dir = infra_dir_path.joinpath("secrets").resolve()
                infra_example_secrets_dir = infra_dir_path.joinpath("example_secrets").resolve()

                print_info(f"Creating {str(infra_secrets_dir)}")
                copytree(
                    str(infra_example_secrets_dir),
                    str(infra_secrets_dir),
                )
            else:
                log_warning("Could not find infra directory - skipping secrets setup")
        except Exception as e:
            log_warning(f"Could not create infra/secrets: {e}")
            log_warning("Please manually copy infra/example_secrets to infra/secrets")

    infra_config = agno_config.create_or_update_infra_config(infra_root_path=infra_root_path, set_as_active=True)

    if infra_config is not None:
        relative_infra_root_path = infra_root_path.relative_to(current_dir)
        print_info("\n--------------------------------")
        print_info(f"Done! Your new AgentOS codebase is available at: ./{str(relative_infra_root_path)} \n")
        print_info(f"Please run `cd ./{str(relative_infra_root_path)}` and:\n")
        print_info("1. Start your AgentOS -> `ag infra up`")
        print_info("2. Stop your AgentOS -> `ag infra down`")
        print_info("\nView your AgentOS on https://os.agno.com")
        print_info("--------------------------------")

        return infra_config
    else:
        print_info("AgentOS codebase setup unsuccessful. Please try again.")
    return None


def setup_infra_config_from_dir(infra_root_path: Path) -> Optional[InfraConfig]:
    from agno.cli.operator import initialize_agno_cli
    from agno.infra.helpers import get_infra_dir_path

    # 1.2 Create AgnoCliConfig if needed
    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return None

    # Check if the infra contains a `infra` dir
    infra_dir_path = get_infra_dir_path(infra_root_path)
    if infra_dir_path is not None:
        logger.debug(f"Found the `infra` configuration at: {infra_dir_path}")
    else:
        logger.debug("No infra directory found, but continuing setup")
    infra_config = agno_config.create_or_update_infra_config(infra_root_path=infra_root_path, set_as_active=True)
    if infra_config is None:
        logger.error(f"Failed to create InfraConfig for {infra_root_path}")
        return None
    return infra_config


def start_infra(
    infra_config: InfraConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
    pull: Optional[bool] = False,
) -> None:
    """Start an AgentOS codebase infrastructure. This is called from `ag infra up`"""

    print_heading("Starting AgentOS codebase infrastructure: {}".format(str(infra_config.infra_root_path.stem)))
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")
    logger.debug(f"\tpull         : {pull}")

    # Set the local environment variables before processing configs
    infra_config.set_local_env()

    # Get resource groups to deploy
    resource_groups_to_create: List[InfraResources] = infra_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups created
    num_rgs_created = 0
    num_rgs_to_create = len(resource_groups_to_create)
    # Track number of resources created
    num_resources_created = 0
    num_resources_to_create = 0

    if num_rgs_to_create == 0:
        print_info("No resources to create")
        return

    logger.debug(f"Deploying {num_rgs_to_create} resource groups")
    for rg in resource_groups_to_create:
        _num_resources_created, _num_resources_to_create = rg.create_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
            pull=pull,
        )
        if _num_resources_created > 0:
            num_rgs_created += 1
        num_resources_created += _num_resources_created
        num_resources_to_create += _num_resources_to_create
        logger.debug(f"Deployed {num_resources_created} resources in {num_rgs_created} resource groups")

    if dry_run:
        return

    if num_resources_created == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deployed: {num_rgs_created}/{num_rgs_to_create}\n")

    if num_resources_created != num_resources_to_create:
        logger.error("Some resources failed to create, please check logs")


def stop_infra(
    infra_config: InfraConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    """Stop an Agno Infra. This is called from `ag infra down`"""
    print_heading("Stopping infra: {}".format(str(infra_config.infra_root_path.stem)))
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    # Set the local environment variables before processing configs
    infra_config.set_local_env()

    # Get resource groups to delete
    resource_groups_to_delete: List[InfraResources] = infra_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="delete",
    )

    # Track number of resource groups deleted
    num_rgs_deleted = 0
    num_rgs_to_delete = len(resource_groups_to_delete)
    # Track number of resources deleted
    num_resources_deleted = 0
    num_resources_to_delete = 0

    if num_rgs_to_delete == 0:
        print_info("No resources to delete")
        return

    logger.debug(f"Deleting {num_rgs_to_delete} resource groups")
    for rg in resource_groups_to_delete:
        _num_resources_deleted, _num_resources_to_delete = rg.delete_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_deleted > 0:
            num_rgs_deleted += 1
        num_resources_deleted += _num_resources_deleted
        num_resources_to_delete += _num_resources_to_delete
        logger.debug(f"Deleted {num_resources_deleted} resources in {num_rgs_deleted} resource groups")

    if dry_run:
        return

    if num_resources_deleted == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deleted: {num_rgs_deleted}/{num_rgs_to_delete}\n")

    if num_resources_to_delete != num_resources_deleted:
        logger.error("Some resources failed to delete, please check logs")


def update_infra(
    infra_config: InfraConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
    pull: Optional[bool] = False,
) -> None:
    """Update an Agno Infra. This is called from `ag infra patch`"""
    print_heading("Updating infra: {}".format(str(infra_config.infra_root_path.stem)))
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")
    logger.debug(f"\tpull         : {pull}")

    # Set the local environment variables before processing configs
    infra_config.set_local_env()

    # Get resource groups to update
    resource_groups_to_update: List[InfraResources] = infra_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="create",
    )
    # Track number of resource groups updated
    num_rgs_updated = 0
    num_rgs_to_update = len(resource_groups_to_update)
    # Track number of resources updated
    num_resources_updated = 0
    num_resources_to_update = 0

    if num_rgs_to_update == 0:
        print_info("No resources to update")
        return

    logger.debug(f"Updating {num_rgs_to_update} resource groups")
    for rg in resource_groups_to_update:
        _num_resources_updated, _num_resources_to_update = rg.update_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
            pull=pull,
        )
        if _num_resources_updated > 0:
            num_rgs_updated += 1
        num_resources_updated += _num_resources_updated
        num_resources_to_update += _num_resources_to_update
        logger.debug(f"Updated {num_resources_updated} resources in {num_rgs_updated} resource groups")

    if dry_run:
        return

    if num_resources_updated == 0:
        return

    print_heading(f"\n--**-- ResourceGroups updated: {num_rgs_updated}/{num_rgs_to_update}\n")

    if num_resources_updated != num_resources_to_update:
        logger.error("Some resources failed to update, please check logs")


def delete_infra(agno_config: AgnoCliConfig, infra_to_delete: Optional[List[Path]]) -> None:
    if infra_to_delete is None or len(infra_to_delete) == 0:
        print_heading("No infra to delete")
        return

    for infra_root in infra_to_delete:
        agno_config.delete_infra(infra_root_path=infra_root)


def set_infra_as_active(infra_dir_name: Optional[str]) -> None:
    from agno.cli.operator import initialize_agno_cli

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check AgnoCliConfig is valid
    ######################################################
    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return

    ######################################################
    # 1.2 Check infra_root_path is valid
    ######################################################
    # By default, we assume this command is run from the infra directory
    if infra_dir_name is None:
        # If the user does not provide a infra_name, that implies `ag set` is ran from
        # the infra directory.
        infra_root_path = Path("").resolve()
    else:
        # If the user provides a infra name manually, we find the dir for that infra
        infra_config: Optional[InfraConfig] = agno_config.get_infra_config_by_dir_name(infra_dir_name)
        if infra_config is None:
            logger.error(f"Could not find infra {infra_dir_name}")
            return
        infra_root_path = infra_config.infra_root_path

    infra_dir_is_valid: bool = infra_root_path is not None and infra_root_path.exists() and infra_root_path.is_dir()
    if not infra_dir_is_valid:
        logger.error("Invalid codebase directory: {}".format(infra_root_path))
        return

    ######################################################
    # 1.3 Validate InfraConfig is available i.e. a Infra is available at this directory
    ######################################################
    logger.debug(f"Checking for an AgentOS codebase at path: {infra_root_path}")
    active_infra_config: Optional[InfraConfig] = agno_config.get_infra_config_by_path(infra_root_path)
    if active_infra_config is None:
        # This happens when the infra is not yet setup
        print_info(f"Could not find a AgentOS codebase at path: {infra_root_path}")
        # TODO: setup automatically for the user
        # print_info("If this AgentOS codebase has not been setup, please run `ag infra setup` from the infra directory")
        return

    ######################################################
    ## 2. Set Infra as active
    ######################################################
    print_heading(f"Setting AgentOS codebase {active_infra_config.infra_root_path.stem} as active")
    agno_config.set_active_infra_dir(active_infra_config.infra_root_path)
    print_info("Active AgentOS codebase updated")
    return



import subprocess
from os import environ
from pathlib import Path
from sys import stderr, stdin, stdout
from typing import List

from agno.utilities.logging import logger


def find_compose_files(directory: Path) -> List[Path]:
    """Find Docker Compose files in the given directory."""
    compose_files = []
    common_names = ["docker-compose.yml", "docker-compose.yaml", "compose.yml", "compose.yaml"]

    for name in common_names:
        compose_file = directory / name
        if compose_file.exists():
            compose_files.append(compose_file)

    return compose_files


def run_docker_compose_up(compose_file: Path, build: bool = True, detached: bool = True, pull: bool = False) -> bool:
    """Run docker compose up command."""
    cmd = ["docker", "compose", "-f", str(compose_file), "up"]

    if detached:
        cmd.append("-d")
    if build:
        cmd.append("--build")
    if pull:
        cmd.append("--pull")
        cmd.append("always")

    try:
        logger.info(f"Running: {' '.join(cmd)}")
        _ = subprocess.run(
            cmd,
            check=True,
            cwd=compose_file.parent,
            env=environ.copy(),
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
        )

        logger.info(f"Docker Compose started successfully from {compose_file}")
        return True

    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to start Docker Compose from {compose_file}")
        logger.error(f"Command: {' '.join(cmd)}")
        logger.error(f"Exit code: {e.returncode}")
        return False
    except FileNotFoundError:
        logger.error("Docker or docker compose command not found. Please ensure Docker is installed and in PATH.")
        return False


def run_docker_compose_down(compose_file: Path, remove_volumes: bool = False, remove_images: bool = False) -> bool:
    """Run docker compose down command."""
    cmd = ["docker", "compose", "-f", str(compose_file), "down"]

    if remove_volumes:
        cmd.append("--volumes")
    if remove_images:
        cmd.append("--rmi")
        cmd.append("all")

    try:
        logger.info(f"Running: {' '.join(cmd)}")
        _ = subprocess.run(
            cmd, check=True, cwd=compose_file.parent, env=environ.copy(), stdin=stdin, stdout=stdout, stderr=stderr
        )

        logger.info(f"Docker Compose stopped successfully from {compose_file}")
        return True

    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to stop Docker Compose from {compose_file}")
        logger.error(f"Command: {' '.join(cmd)}")
        logger.error(f"Exit code: {e.returncode}")
        return False
    except FileNotFoundError:
        logger.error("Docker or docker compose command not found. Please ensure Docker is installed and in PATH.")
        return False



from __future__ import annotations

from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, model_validator


class InfraSettings(BaseModel):
    """Settings that can be used by any resource in the infrastructure."""

    # Infrastructure name
    infra_name: str

    # Path to the infrastructure root
    infra_root: Path

    # infrastructure git repo url
    infra_repo: Optional[str] = None

    # default env for agno infra commands
    default_env: Optional[str] = "dev"

    # default infra for agno infra commands
    default_infra: Optional[str] = None

    # Image Settings
    # Repository for images
    image_repo: str = "agnohq"
    # 'name:tag' for the image
    image_name: Optional[str] = None
    # If True, build images locally
    build_images: bool = False
    # If True, push images after building
    push_images: bool = False
    # If True, skip cache when building images
    skip_image_cache: bool = False
    # If True, force pull images in FROM
    force_pull_images: bool = False

    # Test Settings
    test_env: str = "test"
    test_key: Optional[str] = None

    # Development Settings
    dev_env: str = "dev"
    dev_key: Optional[str] = None

    # Staging Settings
    stg_env: str = "stg"
    stg_key: Optional[str] = None

    # Production Settings
    prd_env: str = "prd"
    prd_key: Optional[str] = None

    # ag cli settings
    # Set to True if Agno should continue creating
    # resources after a resource creation has failed
    continue_on_create_failure: bool = False
    # Set to True if Agno should continue deleting
    # resources after a resource deleting has failed
    # Defaults to True because we normally want to continue deleting
    continue_on_delete_failure: bool = True
    # Set to True if Agno should continue patching
    # resources after a resource patch has failed
    continue_on_patch_failure: bool = False
    # Other Settings
    # Use cached resource if available, i.e. skip resource creation if the resource already exists
    use_cache: bool = True

    ## AWS ##
    # AWS Region and Profile
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None

    # AWS Subnet Configuration
    aws_subnet_ids: List[str] = []
    # Public subnets. Will be added to aws_subnet_ids if provided and aws_subnet_ids is empty.
    # Note: not added to aws_subnet_ids if aws_subnet_ids is provided.
    aws_public_subnets: List[str] = []
    # Private subnets. Will be added to aws_subnet_ids if provided and aws_subnet_ids is empty.
    # Note: not added to aws_subnet_ids if aws_subnet_ids is provided.
    aws_private_subnets: List[str] = []

    # AWS Availability Zones
    aws_az1: Optional[str] = None
    aws_az2: Optional[str] = None
    aws_az3: Optional[str] = None
    aws_az4: Optional[str] = None
    aws_az5: Optional[str] = None

    # AWS Security Groups
    aws_security_group_ids: List[str] = []

    # AWS EFS (Elastic File System) for persistent storage
    efs_file_system_id: Optional[str] = None
    # EFS Access Point ID (for user/permission mapping)
    efs_access_point_id: Optional[str] = None

    @model_validator(mode="after")
    def validate_and_set_computed_fields(self) -> "InfraSettings":
        """Validate and set computed fields after initialization."""
        self._validate_and_set_keys()
        self._validate_and_set_subnet_ids()
        return self

    def _validate_and_set_keys(self):
        """Validate and set environment keys if not provided."""
        if self.test_key is None:
            if self.infra_name is None:
                raise ValueError("`infra_name` is None: Please set a valid value")
            if self.test_env is None:
                raise ValueError("`test_env` is None: Please set a valid value")
            self.test_key = f"{self.infra_name}-{self.test_env}"

        if self.dev_key is None:
            if self.infra_name is None:
                raise ValueError("`infra_name` is None: Please set a valid value")
            if self.dev_env is None:
                raise ValueError("`dev_env` is None: Please set a valid value")
            self.dev_key = f"{self.infra_name}-{self.dev_env}"

        if self.stg_key is None:
            if self.infra_name is None:
                raise ValueError("`infra_name` is None: Please set a valid value")
            if self.stg_env is None:
                raise ValueError("`stg_env` is None: Please set a valid value")
            self.stg_key = f"{self.infra_name}-{self.stg_env}"

        if self.prd_key is None:
            if self.infra_name is None:
                raise ValueError("`infra_name` is None: Please set a valid value")
            if self.prd_env is None:
                raise ValueError("`prd_env` is None: Please set a valid value")
            self.prd_key = f"{self.infra_name}-{self.prd_env}"

    def _validate_and_set_subnet_ids(self):
        """Set subnet IDs from public and private subnets if not provided."""
        if not self.aws_subnet_ids:
            self.aws_subnet_ids = self.aws_public_subnets + self.aws_private_subnets



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.aws.resource.cloudformation.stack import CloudFormationStack
from agno.aws.resource.ec2.security_group import SecurityGroup
from agno.aws.resource.rds.db_instance import DbInstance
from agno.aws.resource.rds.db_subnet_group import DbSubnetGroup
from agno.aws.resource.secret.manager import SecretsManager
from agno.cli.console import print_info
from agno.utilities.logging import logger


class DbCluster(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html
    """

    resource_type: Optional[str] = "DbCluster"
    service_name: str = "rds"

    # Name of the cluster.
    name: str
    # The name of the database engine to be used for this DB cluster.
    engine: Union[str, Literal["aurora", "aurora-mysql", "aurora-postgresql", "mysql", "postgres"]]
    # The version number of the database engine to use.
    # For valid engine_version values, refer to
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html#RDS.Client.create_db_cluster
    # Or run: aws rds describe-db-engine-versions --engine postgres --query "DBEngineVersions[].EngineVersion"
    engine_version: Optional[str] = None
    # DbInstances to add to this cluster
    db_instances: Optional[List[DbInstance]] = None

    # The name for your database of up to 64 alphanumeric characters.
    # If you do not provide a name, Amazon RDS doesn't create a database in the DB cluster you are creating.
    # Provide DATABASE_NAME here or as DATABASE_NAME in secrets_file
    # Valid for: Aurora DB clusters and Multi-AZ DB clusters
    database_name: Optional[str] = None
    # The DB cluster identifier. This parameter is stored as a lowercase string.
    # If None, use name as the db_cluster_identifier
    # Constraints:
    #   Must contain from 1 to 63 letters, numbers, or hyphens.
    #   First character must be a letter.
    #   Can't end with a hyphen or contain two consecutive hyphens.
    # Example: my-cluster1
    # Valid for: Aurora DB clusters and Multi-AZ DB clusters
    db_cluster_identifier: Optional[str] = None
    # The name of the DB cluster parameter group to associate with this DB cluster.
    # If you do not specify a value, then the default DB cluster parameter group for the specified
    # DB engine and version is used.
    # Constraints: If supplied, must match the name of an existing DB cluster parameter group.
    db_cluster_parameter_group_name: Optional[str] = None

    # The port number on which the instances in the DB cluster accept connections.
    # RDS for MySQL and Aurora MySQL
    # - Default: 3306
    # - Valid values: 1150-65535
    # RDS for PostgreSQL and Aurora PostgreSQL
    # - Default: 5432
    # - Valid values: 1150-65535
    port: Optional[int] = None

    # The name of the master user for the DB cluster.
    # Constraints:
    #   Must be 1 to 16 letters or numbers.
    #   First character must be a letter.
    #   Can't be a reserved word for the chosen database engine.
    # Provide MASTER_USERNAME here or as MASTER_USERNAME in secrets_file
    master_username: Optional[str] = None
    # The password for the master database user. This password can contain any printable ASCII character
    # except "/", """, or "@".
    # Constraints: Must contain from 8 to 41 characters.
    # Provide MASTER_USER_PASSWORD here or as MASTER_USER_PASSWORD in secrets_file
    master_user_password: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secret
    aws_secret: Optional[SecretsManager] = None

    # A list of Availability Zones (AZs) where DB instances in the DB cluster can be created.
    # Valid for: Aurora DB clusters only
    availability_zones: Optional[List[str]] = None
    # A DB subnet group to associate with this DB cluster.
    # This setting is required to create a Multi-AZ DB cluster.
    # Constraints: Must match the name of an existing DBSubnetGroup. Must not be default.
    db_subnet_group_name: Optional[str] = None
    # If db_subnet_group_name is None,
    # Read the db_subnet_group_name from db_subnet_group
    db_subnet_group: Optional[DbSubnetGroup] = None

    # Compute and memory capacity of each DB instance in the Multi-AZ DB cluster, for example db.m6g.xlarge.
    # Not all DB instance classes are available in all Amazon Web Services Regions, or for all database engines.
    # This setting is required to create a Multi-AZ DB cluster.
    db_instance_class: Optional[str] = None
    # The amount of storage in gibibytes (GiB) to allocate to each DB instance in the Multi-AZ DB cluster.
    allocated_storage: Optional[int] = None
    # The storage type to associate with the DB cluster.
    # This setting is required to create a Multi-AZ DB cluster.
    # When specified for a Multi-AZ DB cluster, a value for the Iops parameter is required.
    # Valid Values:
    # Aurora DB clusters - aurora | aurora-iopt1
    # Multi-AZ DB clusters - io1
    # Default:
    # Aurora DB clusters - aurora
    # Multi-AZ DB clusters - io1
    storage_type: Optional[str] = None
    # The amount of Provisioned IOPS (input/output operations per second) to be initially allocated for each DB
    # instance in the Multi-AZ DB cluster.
    iops: Optional[int] = None
    # Specifies whether the DB cluster is publicly accessible.
    # When the DB cluster is publicly accessible, its Domain Name System (DNS) endpoint resolves to the private
    #   IP address from within the DB cluster’s virtual private cloud (VPC). It resolves to the public IP address
    #   from outside the DB cluster’s VPC. Access to the DB cluster is ultimately controlled by the security group
    #   it uses. That public access isn’t permitted if the security group assigned to the DB cluster doesn’t permit it.
    # When the DB cluster isn’t publicly accessible, it is an internal DB cluster with a DNS name that resolves to a
    #   private IP address.
    # Default: The default behavior varies depending on whether DBSubnetGroupName is specified.
    # If DBSubnetGroupName isn’t specified, and PubliclyAccessible isn’t specified, the following applies:
    # If the default VPC in the target Region doesn’t have an internet gateway attached to it, the DB cluster is private
    # If the default VPC in the target Region has an internet gateway attached to it, the DB cluster is public.
    # If DBSubnetGroupName is specified, and PubliclyAccessible isn’t specified, the following applies:
    # If the subnets are part of a VPC that doesn’t have an internet gateway attached to it, the DB cluster is private.
    # If the subnets are part of a VPC that has an internet gateway attached to it, the DB cluster is public.
    publicly_accessible: Optional[bool] = None

    # A list of EC2 VPC security groups to associate with this DB cluster.
    vpc_security_group_ids: Optional[List[str]] = None
    # If vpc_security_group_ids is None,
    # Read the security_group_id from vpc_stack
    vpc_stack: Optional[CloudFormationStack] = None
    # Add security_group_ids from db_security_groups
    db_security_groups: Optional[List[SecurityGroup]] = None

    # The DB engine mode of the DB cluster, either provisioned or serverless
    engine_mode: Optional[Literal["provisioned", "serverless"]] = None
    # For DB clusters in serverless DB engine mode, the scaling properties of the DB cluster.
    scaling_configuration: Optional[Dict[str, Any]] = None
    # Contains the scaling configuration of an Aurora Serverless v2 DB cluster.
    serverless_v2_scaling_configuration: Dict[str, Any] = {
        "MinCapacity": 0.5,
        "MaxCapacity": 8,
    }
    # A value that indicates whether the DB cluster has deletion protection enabled.
    # The database can't be deleted when deletion protection is enabled. By default, deletion protection isn't enabled.
    deletion_protection: Optional[bool] = None

    # The number of days for which automated backups are retained.
    # Default: 1
    # Constraints: Must be a value from 1 to 35
    # Valid for: Aurora DB clusters and Multi-AZ DB clusters
    backup_retention_period: Optional[int] = None
    # A value that indicates that the DB cluster should be associated with the specified CharacterSet.
    # Valid for: Aurora DB clusters only
    character_set_name: Optional[str] = None

    # A value that indicates that the DB cluster should be associated with the specified option group.
    option_group_name: Optional[str] = None
    # The daily time range during which automated backups are created if automated backups are enabled
    # using the BackupRetentionPeriod parameter.
    # The default is a 30-minute window selected at random from an 8-hour block of time for each
    # Amazon Web Services Region.
    # Constraints:
    #   Must be in the format hh24:mi-hh24:mi .
    #   Must be in Universal Coordinated Time (UTC).
    #   Must not conflict with the preferred maintenance window.
    #   Must be at least 30 minutes.
    preferred_backup_window: Optional[str] = None
    # The weekly time range during which system maintenance can occur, in Universal Coordinated Time (UTC).
    # Format: ddd:hh24:mi-ddd:hh24:mi
    # The default is a 30-minute window selected at random from an 8-hour block of time for each
    # Amazon Web Services Region, occurring on a random day of the week.
    # Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun.
    # Constraints: Minimum 30-minute window.
    preferred_maintenance_window: Optional[str] = None
    # The Amazon Resource Name (ARN) of the source DB instance or DB cluster
    # if this DB cluster is created as a read replica.
    replication_source_identifier: Optional[str] = None
    # Tags to assign to the DB cluster.
    tags: Optional[List[Dict[str, str]]] = None
    # A value that indicates whether the DB cluster is encrypted.
    storage_encrypted: Optional[bool] = None
    # The Amazon Web Services KMS key identifier for an encrypted DB cluster.
    kms_key_id: Optional[str] = None
    pre_signed_url: Optional[str] = None
    # A value that indicates whether to enable mapping of Amazon Web Services Identity and Access Management (IAM)
    # accounts to database accounts. By default, mapping isn't enabled.
    enable_iam_database_authentication: Optional[bool] = None
    # The target backtrack window, in seconds. To disable backtracking, set this value to 0.
    # Default: 0
    backtrack_window: Optional[int] = None
    # The list of log types that need to be enabled for exporting to CloudWatch Logs.
    # The values in the list depend on the DB engine being used.
    # RDS for MySQL: Possible values are error , general , and slowquery .
    # RDS for PostgreSQL: Possible values are postgresql and upgrade .
    # Aurora MySQL: Possible values are audit , error , general , and slowquery .
    # Aurora PostgreSQL: Possible value is postgresql .
    enable_cloudwatch_logs_exports: Optional[List[str]] = None

    # The global cluster ID of an Aurora cluster that becomes the primary cluster in the new global database cluster.
    global_cluster_identifier: Optional[str] = None
    # A value that indicates whether to enable the HTTP endpoint for an Aurora Serverless v1 DB cluster.
    # By default, the HTTP endpoint is disabled.
    # When enabled, the HTTP endpoint provides a connectionless web service
    # API for running SQL queries on the Aurora Serverless v1 DB cluster.
    # You can also query your database from inside the RDS console with the query editor.
    enable_http_endpoint: Optional[bool] = None
    # A value that indicates whether to copy all tags from the DB cluster to snapshots of the DB cluster.
    # The default is not to copy them.
    copy_tags_to_snapshot: Optional[bool] = None
    # The Active Directory directory ID to create the DB cluster in.
    # For Amazon Aurora DB clusters, Amazon RDS can use Kerberos authentication to authenticate users that connect to
    # the DB cluster.
    domain: Optional[str] = None
    # Specify the name of the IAM role to be used when making API calls to the Directory Service.
    domain_iam_role_name: Optional[str] = None
    enable_global_write_forwarding: Optional[bool] = None

    # A value that indicates whether minor engine upgrades are applied automatically to the DB cluster during the
    # maintenance window. By default, minor engine upgrades are applied automatically.
    auto_minor_version_upgrade: Optional[bool] = None
    # The interval, in seconds, between points when Enhanced Monitoring metrics are collected for the DB cluster.
    # To turn off collecting Enhanced Monitoring metrics, specify 0. The default is 0.
    # If MonitoringRoleArn is specified, also set MonitoringInterval to a value other than 0.
    # Valid Values: 0, 1, 5, 10, 15, 30, 60
    monitoring_interval: Optional[int] = None
    # The Amazon Resource Name (ARN) for the IAM role that permits RDS to send
    # Enhanced Monitoring metrics to Amazon CloudWatch Logs.
    monitoring_role_arn: Optional[str] = None
    enable_performance_insights: Optional[bool] = None
    performance_insights_kms_key_id: Optional[str] = None
    performance_insights_retention_period: Optional[int] = None

    # The network type of the DB cluster.
    # Valid values:
    # - IPV4
    # - DUAL
    # The network type is determined by the DBSubnetGroup specified for the DB cluster.
    # A DBSubnetGroup can support only the IPv4 protocol or the IPv4 and the IPv6 protocols (DUAL ).
    network_type: Optional[str] = None
    # Reserved for future use.
    db_system_id: Optional[str] = None
    # The ID of the region that contains the source for the db cluster.
    source_region: Optional[str] = None
    enable_local_write_forwarding: Optional[bool] = None

    # Specifies whether to manage the master user password with Amazon Web Services Secrets Manager.
    # Constraints:
    # Can’t manage the master user password with Amazon Web Services Secrets Manager if MasterUserPassword is specified.
    manage_master_user_password: Optional[bool] = None
    # The Amazon Web Services KMS key identifier to encrypt a secret that is automatically generated and
    # managed in Amazon Web Services Secrets Manager.
    master_user_secret_kms_key_id: Optional[str] = None

    # Parameters for delete function
    # Skip the creation of a final DB cluster snapshot before the DB cluster is deleted.
    # If skip_final_snapshot = True, no DB cluster snapshot is created.
    # If skip_final_snapshot = None | False, a DB cluster snapshot is created before the DB cluster is deleted.
    #   You must specify a FinalDBSnapshotIdentifier parameter
    #   if skip_final_snapshot = None | False
    skip_final_snapshot: Optional[bool] = True
    # The DB cluster snapshot identifier of the new DB cluster snapshot created when SkipFinalSnapshot is disabled.
    final_db_snapshot_identifier: Optional[str] = None
    # Specifies whether to remove automated backups immediately after the DB cluster is deleted.
    # The default is to remove automated backups immediately after the DB cluster is deleted.
    delete_automated_backups: Optional[bool] = None

    # Parameters for update function
    new_db_cluster_identifier: Optional[str] = None
    apply_immediately: Optional[bool] = None
    cloudwatch_logs_exports: Optional[List[str]] = None
    allow_major_version_upgrade: Optional[bool] = None
    db_instance_parameter_group_name: Optional[str] = None
    rotate_master_user_password: Optional[bool] = None
    allow_engine_mode_change: Optional[bool] = None

    # Cache secret_data
    cached_secret_data: Optional[Dict[str, Any]] = None

    def get_db_cluster_identifier(self):
        return self.db_cluster_identifier or self.name

    def get_master_username(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        master_username = self.master_username
        if master_username is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                master_username = secret_data.get("MASTER_USERNAME", master_username)
        if master_username is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading MASTER_USERNAME from secret: {self.aws_secret.name}")
            master_username = self.aws_secret.get_secret_value("MASTER_USERNAME", aws_client=aws_client)

        return master_username

    def get_master_user_password(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        master_user_password = self.master_user_password
        if master_user_password is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                master_user_password = secret_data.get("MASTER_USER_PASSWORD", master_user_password)
        if master_user_password is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading MASTER_USER_PASSWORD from secret: {self.aws_secret.name}")
            master_user_password = self.aws_secret.get_secret_value("MASTER_USER_PASSWORD", aws_client=aws_client)

        return master_user_password

    def get_database_name(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        database_name = self.database_name
        if database_name is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                database_name = secret_data.get("DATABASE_NAME", database_name)
                if database_name is None:
                    database_name = secret_data.get("DB_NAME", database_name)
        if database_name is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading DATABASE_NAME from secret: {self.aws_secret.name}")
            database_name = self.aws_secret.get_secret_value("DATABASE_NAME", aws_client=aws_client)
            if database_name is None:
                logger.debug(f"Reading DB_NAME from secret: {self.aws_secret.name}")
                database_name = self.aws_secret.get_secret_value("DB_NAME", aws_client=aws_client)
        return database_name

    def get_db_name(self) -> Optional[str]:
        # Alias for get_database_name because db_instances use `db_name` and db_clusters use `database_name`
        return self.get_database_name()

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}

        # Step 1: Get the VpcSecurityGroupIds
        vpc_security_group_ids = self.vpc_security_group_ids
        if vpc_security_group_ids is None and self.vpc_stack is not None:
            vpc_stack_sg = self.vpc_stack.get_security_group(aws_client=aws_client)
            if vpc_stack_sg is not None:
                vpc_security_group_ids = [vpc_stack_sg]
        if self.db_security_groups is not None:
            sg_ids = []
            for sg in self.db_security_groups:
                sg_id = sg.get_security_group_id(aws_client)
                if sg_id is not None:
                    sg_ids.append(sg_id)
            if len(sg_ids) > 0:
                if vpc_security_group_ids is None:
                    vpc_security_group_ids = []
                vpc_security_group_ids.extend(sg_ids)
        if vpc_security_group_ids is not None:
            logger.debug(f"Using SecurityGroups: {vpc_security_group_ids}")
            not_null_args["VpcSecurityGroupIds"] = vpc_security_group_ids

        # Step 2: Get the DbSubnetGroupName
        db_subnet_group_name = self.db_subnet_group_name
        if db_subnet_group_name is None and self.db_subnet_group is not None:
            db_subnet_group_name = self.db_subnet_group.name
            logger.debug(f"Using DbSubnetGroup: {db_subnet_group_name}")
        if db_subnet_group_name is not None:
            not_null_args["DBSubnetGroupName"] = db_subnet_group_name

        database_name = self.get_database_name()
        if database_name:
            not_null_args["DatabaseName"] = database_name

        master_username = self.get_master_username()
        if master_username:
            not_null_args["MasterUsername"] = master_username
        master_user_password = self.get_master_user_password()
        if master_user_password:
            not_null_args["MasterUserPassword"] = master_user_password

        if self.availability_zones:
            not_null_args["AvailabilityZones"] = self.availability_zones
        if self.backup_retention_period:
            not_null_args["BackupRetentionPeriod"] = self.backup_retention_period
        if self.character_set_name:
            not_null_args["CharacterSetName"] = self.character_set_name

        if self.db_cluster_parameter_group_name:
            not_null_args["DBClusterParameterGroupName"] = self.db_cluster_parameter_group_name

        if self.engine_version:
            not_null_args["EngineVersion"] = self.engine_version
        if self.port:
            not_null_args["Port"] = self.port

        if self.option_group_name:
            not_null_args["OptionGroupName"] = self.option_group_name
        if self.preferred_backup_window:
            not_null_args["PreferredBackupWindow"] = self.preferred_backup_window
        if self.preferred_maintenance_window:
            not_null_args["PreferredMaintenanceWindow"] = self.preferred_maintenance_window
        if self.replication_source_identifier:
            not_null_args["ReplicationSourceIdentifier"] = self.replication_source_identifier
        if self.tags:
            not_null_args["Tags"] = self.tags
        if self.storage_encrypted:
            not_null_args["StorageEncrypted"] = self.storage_encrypted
        if self.kms_key_id:
            not_null_args["KmsKeyId"] = self.kms_key_id
        if self.enable_iam_database_authentication:
            not_null_args["EnableIAMDbClusterAuthentication"] = self.enable_iam_database_authentication
        if self.backtrack_window:
            not_null_args["BacktrackWindow"] = self.backtrack_window
        if self.enable_cloudwatch_logs_exports:
            not_null_args["EnableCloudwatchLogsExports"] = self.enable_cloudwatch_logs_exports
        if self.engine_mode:
            not_null_args["EngineMode"] = self.engine_mode
        if self.scaling_configuration:
            not_null_args["ScalingConfiguration"] = self.scaling_configuration
        if self.deletion_protection:
            not_null_args["DeletionProtection"] = self.deletion_protection
        if self.global_cluster_identifier:
            not_null_args["GlobalClusterIdentifier"] = self.global_cluster_identifier
        if self.enable_http_endpoint:
            not_null_args["EnableHttpEndpoint"] = self.enable_http_endpoint
        if self.copy_tags_to_snapshot:
            not_null_args["CopyTagsToSnapshot"] = self.copy_tags_to_snapshot
        if self.domain:
            not_null_args["Domain"] = self.domain
        if self.domain_iam_role_name:
            not_null_args["DomainIAMRoleName"] = self.domain_iam_role_name
        if self.enable_global_write_forwarding:
            not_null_args["EnableGlobalWriteForwarding"] = self.enable_global_write_forwarding
        if self.db_instance_class:
            not_null_args["DBClusterInstanceClass"] = self.db_instance_class
        if self.allocated_storage:
            not_null_args["AllocatedStorage"] = self.allocated_storage
        if self.storage_type:
            not_null_args["StorageType"] = self.storage_type
        if self.iops:
            not_null_args["Iops"] = self.iops
        if self.publicly_accessible:
            not_null_args["PubliclyAccessible"] = self.publicly_accessible
        if self.auto_minor_version_upgrade:
            not_null_args["AutoMinorVersionUpgrade"] = self.auto_minor_version_upgrade
        if self.monitoring_interval:
            not_null_args["MonitoringInterval"] = self.monitoring_interval
        if self.monitoring_role_arn:
            not_null_args["MonitoringRoleArn"] = self.monitoring_role_arn
        if self.enable_performance_insights:
            not_null_args["EnablePerformanceInsights"] = self.enable_performance_insights
        if self.performance_insights_kms_key_id:
            not_null_args["PerformanceInsightsKMSKeyId"] = self.performance_insights_kms_key_id
        if self.performance_insights_retention_period:
            not_null_args["PerformanceInsightsRetentionPeriod"] = self.performance_insights_retention_period
        if self.serverless_v2_scaling_configuration:
            not_null_args["ServerlessV2ScalingConfiguration"] = self.serverless_v2_scaling_configuration
        if self.network_type:
            not_null_args["NetworkType"] = self.network_type
        if self.db_system_id:
            not_null_args["DBSystemId"] = self.db_system_id
        if self.source_region:
            not_null_args["SourceRegion"] = self.source_region
        if self.enable_local_write_forwarding:
            not_null_args["EnableLocalWriteForwarding"] = self.enable_local_write_forwarding

        if self.manage_master_user_password:
            not_null_args["ManageMasterUserPassword"] = self.manage_master_user_password
        if self.master_user_secret_kms_key_id:
            not_null_args["MasterUserSecretKmsKeyId"] = self.master_user_secret_kms_key_id

        # Step 3: Create DBCluster
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_db_cluster(
                DBClusterIdentifier=self.get_db_cluster_identifier(),
                Engine=self.engine,
                **not_null_args,
            )
            logger.debug(f"Response: {create_response}")
            resource_dict = create_response.get("DBCluster", {})

            # Validate database creation
            if resource_dict is not None:
                logger.debug(f"DBCluster created: {self.get_db_cluster_identifier()}")
                self.active_resource = resource_dict
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        db_instances_created = []
        if self.db_instances is not None:
            for db_instance in self.db_instances:
                db_instance.db_cluster_identifier = self.get_db_cluster_identifier()
                if db_instance._create(aws_client):  # type: ignore
                    db_instances_created.append(db_instance)

        # Wait for DBCluster to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be active.")
                waiter = self.get_service_client(aws_client).get_waiter("db_cluster_available")
                waiter.wait(
                    DBClusterIdentifier=self.get_db_cluster_identifier(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)

            # Wait for DbInstances to be created
            for db_instance in db_instances_created:
                db_instance.post_create(aws_client)

        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            resource_identifier = self.get_db_cluster_identifier()
            describe_response = service_client.describe_db_clusters(DBClusterIdentifier=resource_identifier)
            logger.debug(f"DbCluster: {describe_response}")
            resources_list = describe_response.get("DBClusters", None)

            if resources_list is not None and isinstance(resources_list, list):
                for _resource in resources_list:
                    _identifier = _resource.get("DBClusterIdentifier", None)
                    if _identifier == resource_identifier:
                        self.active_resource = _resource
                        break
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        # Step 1: Delete DbInstances
        if self.db_instances is not None:
            for db_instance in self.db_instances:
                db_instance._delete(aws_client)

        # Step 2: Delete DbCluster
        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.final_db_snapshot_identifier:
            not_null_args["FinalDBSnapshotIdentifier"] = self.final_db_snapshot_identifier
        if self.delete_automated_backups:
            not_null_args["DeleteAutomatedBackups"] = self.delete_automated_backups

        try:
            db_cluster_identifier = self.get_db_cluster_identifier()
            delete_response = service_client.delete_db_cluster(
                DBClusterIdentifier=db_cluster_identifier,
                SkipFinalSnapshot=self.skip_final_snapshot,
                **not_null_args,
            )
            logger.debug(f"Response: {delete_response}")
            resource_dict = delete_response.get("DBCluster", {})

            # Validate database deletion
            if resource_dict is not None:
                logger.debug(f"DBCluster deleted: {self.get_db_cluster_identifier()}")
                self.active_resource = resource_dict
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for DbCluster to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("db_cluster_deleted")
                waiter.wait(
                    DBClusterIdentifier=self.get_db_cluster_identifier(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Updates the DbCluster"""

        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get existing DBInstance
        db_cluster = self.read(aws_client)

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {
            "ApplyImmediately": self.apply_immediately,
        }

        vpc_security_group_ids = self.vpc_security_group_ids
        if vpc_security_group_ids is None and self.vpc_stack is not None:
            vpc_stack_sg = self.vpc_stack.get_security_group(aws_client=aws_client)
            if vpc_stack_sg is not None:
                vpc_security_group_ids = [vpc_stack_sg]
        if self.db_security_groups is not None:
            sg_ids = []
            for sg in self.db_security_groups:
                sg_id = sg.get_security_group_id(aws_client)
                if sg_id is not None:
                    sg_ids.append(sg_id)
            if len(sg_ids) > 0:
                if vpc_security_group_ids is None:
                    vpc_security_group_ids = []
                vpc_security_group_ids.extend(sg_ids)
        # Check if vpc_security_group_ids has changed
        existing_vpc_security_group = db_cluster.get("VpcSecurityGroups", [])
        existing_vpc_security_group_ids = []
        for existing_sg in existing_vpc_security_group:
            existing_vpc_security_group_ids.append(existing_sg.get("VpcSecurityGroupId", None))
        if vpc_security_group_ids is not None and vpc_security_group_ids != existing_vpc_security_group_ids:
            logger.info(f"Updating SecurityGroups: {vpc_security_group_ids}")
            not_null_args["VpcSecurityGroupIds"] = vpc_security_group_ids

        master_user_password = self.get_master_user_password()
        if master_user_password:
            not_null_args["MasterUserPassword"] = master_user_password

        if self.new_db_cluster_identifier:
            not_null_args["NewDBClusterIdentifier"] = self.new_db_cluster_identifier
        if self.backup_retention_period:
            not_null_args["BackupRetentionPeriod"] = self.backup_retention_period
        if self.db_cluster_parameter_group_name:
            not_null_args["DBClusterParameterGroupName"] = self.db_cluster_parameter_group_name
        if self.port:
            not_null_args["Port"] = self.port

        if self.option_group_name:
            not_null_args["OptionGroupName"] = self.option_group_name
        if self.preferred_backup_window:
            not_null_args["PreferredBackupWindow"] = self.preferred_backup_window
        if self.preferred_maintenance_window:
            not_null_args["PreferredMaintenanceWindow"] = self.preferred_maintenance_window
        if self.enable_iam_database_authentication:
            not_null_args["EnableIAMDbClusterAuthentication"] = self.enable_iam_database_authentication
        if self.backtrack_window:
            not_null_args["BacktrackWindow"] = self.backtrack_window
        if self.cloudwatch_logs_exports:
            not_null_args["CloudwatchLogsExportConfiguration"] = self.cloudwatch_logs_exports
        if self.engine_version:
            not_null_args["EngineVersion"] = self.engine_version
        if self.allow_major_version_upgrade:
            not_null_args["AllowMajorVersionUpgrade"] = self.allow_major_version_upgrade
        if self.db_instance_parameter_group_name:
            not_null_args["DBInstanceParameterGroupName"] = self.db_instance_parameter_group_name
        if self.domain:
            not_null_args["Domain"] = self.domain
        if self.domain_iam_role_name:
            not_null_args["DomainIAMRoleName"] = self.domain_iam_role_name
        if self.scaling_configuration:
            not_null_args["ScalingConfiguration"] = self.scaling_configuration
        if self.deletion_protection:
            not_null_args["DeletionProtection"] = self.deletion_protection
        if self.enable_http_endpoint:
            not_null_args["EnableHttpEndpoint"] = self.enable_http_endpoint
        if self.copy_tags_to_snapshot:
            not_null_args["CopyTagsToSnapshot"] = self.copy_tags_to_snapshot
        if self.enable_global_write_forwarding:
            not_null_args["EnableGlobalWriteForwarding"] = self.enable_global_write_forwarding
        if self.db_instance_class:
            not_null_args["DBClusterInstanceClass"] = self.db_instance_class
        if self.allocated_storage:
            not_null_args["AllocatedStorage"] = self.allocated_storage
        if self.storage_type:
            not_null_args["StorageType"] = self.storage_type
        if self.iops:
            not_null_args["Iops"] = self.iops
        if self.auto_minor_version_upgrade:
            not_null_args["AutoMinorVersionUpgrade"] = self.auto_minor_version_upgrade
        if self.monitoring_interval:
            not_null_args["MonitoringInterval"] = self.monitoring_interval
        if self.monitoring_role_arn:
            not_null_args["MonitoringRoleArn"] = self.monitoring_role_arn
        if self.enable_performance_insights:
            not_null_args["EnablePerformanceInsights"] = self.enable_performance_insights
        if self.performance_insights_kms_key_id:
            not_null_args["PerformanceInsightsKMSKeyId"] = self.performance_insights_kms_key_id
        if self.performance_insights_retention_period:
            not_null_args["PerformanceInsightsRetentionPeriod"] = self.performance_insights_retention_period
        if self.serverless_v2_scaling_configuration:
            not_null_args["ServerlessV2ScalingConfiguration"] = self.serverless_v2_scaling_configuration
        if self.network_type:
            not_null_args["NetworkType"] = self.network_type
        if self.manage_master_user_password:
            not_null_args["ManageMasterUserPassword"] = self.manage_master_user_password
        if self.rotate_master_user_password:
            not_null_args["RotateMasterUserPassword"] = self.rotate_master_user_password
        if self.master_user_secret_kms_key_id:
            not_null_args["MasterUserSecretKmsKeyId"] = self.master_user_secret_kms_key_id
        if self.engine_mode:
            not_null_args["EngineMode"] = self.engine_mode
        if self.allow_engine_mode_change:
            not_null_args["AllowEngineModeChange"] = self.allow_engine_mode_change

        # Step 2: Update DBCluster
        service_client = self.get_service_client(aws_client)
        try:
            update_response = service_client.modify_db_cluster(
                DBClusterIdentifier=self.get_db_cluster_identifier(),
                **not_null_args,
            )
            logger.debug(f"Response: {update_response}")
            resource_dict = update_response.get("DBCluster", {})

            # Validate resource update
            if resource_dict is not None:
                print_info(f"DBCluster updated: {self.get_resource_name()}")
                self.active_resource = update_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be updated.")
            logger.error(e)
        return False

    def get_db_endpoint(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the DbCluster endpoint

        Returns:
            The DbCluster endpoint
        """
        logger.debug(f"Getting endpoint for {self.get_resource_name()}")
        _db_endpoint: Optional[str] = None
        if self.active_resource:
            _db_endpoint = self.active_resource.get("Endpoint")
        if _db_endpoint is None:
            client: AwsApiClient = aws_client or self.get_aws_client()
            resource = self._read(aws_client=client)
            if resource is not None:
                _db_endpoint = resource.get("Endpoint")
        if _db_endpoint is None:
            resource = self.read_resource_from_file()
            if resource is not None:
                _db_endpoint = resource.get("Endpoint")
        logger.debug(f"DBCluster Endpoint: {_db_endpoint}")
        return _db_endpoint

    def get_db_reader_endpoint(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the DbCluster reader endpoint

        Returns:
            The DbCluster reader endpoint
        """
        logger.debug(f"Getting endpoint for {self.get_resource_name()}")
        _db_endpoint: Optional[str] = None
        if self.active_resource:
            _db_endpoint = self.active_resource.get("ReaderEndpoint")
        if _db_endpoint is None:
            client: AwsApiClient = aws_client or self.get_aws_client()
            resource = self._read(aws_client=client)
            if resource is not None:
                _db_endpoint = resource.get("ReaderEndpoint")
        if _db_endpoint is None:
            resource = self.read_resource_from_file()
            if resource is not None:
                _db_endpoint = resource.get("ReaderEndpoint")
        logger.debug(f"DBCluster ReaderEndpoint: {_db_endpoint}")
        return _db_endpoint

    def get_db_port(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the DbCluster port

        Returns:
            The DbCluster port
        """
        logger.debug(f"Getting port for {self.get_resource_name()}")
        _db_port: Optional[str] = None
        if self.active_resource:
            _db_port = self.active_resource.get("Port")
        if _db_port is None:
            client: AwsApiClient = aws_client or self.get_aws_client()
            resource = self._read(aws_client=client)
            if resource is not None:
                _db_port = resource.get("Port")
        if _db_port is None:
            resource = self.read_resource_from_file()
            if resource is not None:
                _db_port = resource.get("Port")
        logger.debug(f"DBCluster Port: {_db_port}")
        return _db_port



from __future__ import annotations

from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict

AGNO_CLI_CONFIG_DIR: Path = Path.home().resolve().joinpath(".config").joinpath("ag")


class AgnoCliSettings(BaseSettings):
    config_file_path: Path = AGNO_CLI_CONFIG_DIR.joinpath("config.json")

    model_config = SettingsConfigDict(env_prefix="AGNO_")


agno_cli_settings = AgnoCliSettings()



from agno.base.base import InfraBase
from agno.base.context import ContainerContext
from agno.base.db_app import InfraApp
from agno.base.resource import InfraResource
from agno.base.resources import InfraResources

__all__ = ["InfraBase", "ContainerContext", "InfraApp", "InfraResource", "InfraResources"]



from pathlib import Path
from typing import Dict, Optional

from agno.utilities.logging import log_debug, logger


def read_pyproject_agno(pyproject_file: Path) -> Optional[Dict]:
    """
    Read the pyproject.toml file and return the "agno" tool configuration.
    """
    log_debug(f"Reading {pyproject_file}")
    try:
        # Try tomllib first (Python 3.11+), then fall back to tomli
        try:
            import tomllib

            with open(pyproject_file, "rb") as f:
                pyproject_dict = tomllib.load(f)
        except ImportError:
            import tomli  # type: ignore

            pyproject_dict = tomli.loads(pyproject_file.read_text())

        agno_conf = pyproject_dict.get("tool", {}).get("agno", None)
        if agno_conf is not None and isinstance(agno_conf, dict):
            return agno_conf
    except Exception as e:
        logger.error(f"Could not read {pyproject_file}: {e}")
    return None



from typing import List, Optional, Union

from agno.docker.app.base import ContainerContext, DockerApp  # noqa: F401


class CeleryWorker(DockerApp):
    # -*- App Name
    name: str = "celery-worker"

    # -*- Image Configuration
    image_name: str = "agnohq/celery-worker"
    image_tag: str = "latest"
    command: Optional[Union[str, List[str]]] = "celery -A tasks.celery worker --loglevel=info"

    # -*- OS Configuration
    # Path to the os directory inside the container
    infra_dir_container_path: str = "/app"
    # Mount the infra directory from host machine to the container
    mount_infra: bool = False



from pathlib import Path
from typing import Optional

from agno.utilities.logging import log_debug


def get_infra_dir_from_env() -> Optional[Path]:
    from os import getenv

    from agno.constants import AGNO_INFRA_DIR

    log_debug(f"Reading {AGNO_INFRA_DIR} from environment variables")
    infra_dir = getenv(AGNO_INFRA_DIR, None)
    if infra_dir is not None:
        return Path(infra_dir)
    return None


def is_docker_only_project(infra_root_path: Path) -> bool:
    """
    Check if the project is a docker only project.
    """

    infrastructure_indicators = [
        "compose.yaml",
        "docker-compose.yaml",
        "docker-compose.yml",
        "Dockerfile",
        "compose.yml",
    ]

    for indicator in infrastructure_indicators:
        if infra_root_path.joinpath(indicator).exists():
            log_debug(f"Found infrastructure file '{indicator}' - using root directory as infra path")
            return True
    return False


def get_infra_dir_path(infra_root_path: Path) -> Optional[Path]:
    """
    Get the infra directory path from the given project root path.
    Agno infra dir can be found at:
        1. subdirectory: infra
        2. In a folder defined by the pyproject.toml file
        3. Fallback to root directory if it contains infrastructure files
        4. Returns None if not found (instead of exiting)
    """
    from agno.utilities.pyproject import read_pyproject_agno

    log_debug(f"Searching for a infra directory in {infra_root_path}")

    # Case 1: Look for a subdirectory with name: infra
    infra_dir = infra_root_path.joinpath("infra")
    log_debug(f"Searching {infra_dir}")
    if infra_dir.exists() and infra_dir.is_dir():
        return infra_dir

    # Case 2: Look for a folder defined by the pyproject.toml file
    pyproject_toml_path = infra_root_path.joinpath("pyproject.toml")
    if pyproject_toml_path.exists() and pyproject_toml_path.is_file():
        agno_conf = read_pyproject_agno(pyproject_toml_path)
        if agno_conf is not None:
            agno_conf_infra_dir_str = agno_conf.get("infra-path", None)
            if agno_conf_infra_dir_str is not None:
                agno_conf_infra_dir_path = infra_root_path.joinpath(agno_conf_infra_dir_str)
                log_debug(f"Searching {agno_conf_infra_dir_path}")
                if agno_conf_infra_dir_path.exists() and agno_conf_infra_dir_path.is_dir():
                    return agno_conf_infra_dir_path

    log_debug(f"Could not find a infra directory at: {infra_root_path}")
    return None
